import pathlib
from collections import defaultdict
from typing import Any, TypedDict, cast

import pydantic_core
import tqdm
from eth_typing import ABIEvent, ChecksumAddress, HexAddress, HexStr
from eth_utils.abi import event_abi_to_log_topic
from hexbytes import HexBytes
from web3 import Web3
from web3.contract.base_contract import BaseContractEvent
from web3.types import EventData, LogReceipt
from web3.utils import get_abi_element

from degenbot import connection_manager, get_checksum_address
from degenbot.functions import fetch_logs_retrying
from degenbot.logging import logger
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.types.concrete import KeyedDefaultDict
from degenbot.uniswap.abi import UNISWAP_V4_POOL_MANAGER_ABI
from degenbot.uniswap.v4_types import (
    UniswapV4BitmapAtWord,
    UniswapV4LiquidityAtTick,
    UniswapV4LiquidityEvent,
    UniswapV4PoolLiquidityMappingUpdate,
)

type PoolManagerAddress = ChecksumAddress
type PoolId = str
type ManagedPoolIdentifier = tuple[PoolManagerAddress, PoolId]


class LiquidityMap(TypedDict):
    tick_bitmap: dict[int, UniswapV4BitmapAtWord]
    tick_data: dict[int, UniswapV4LiquidityAtTick]


class UniswapV4LiquiditySnapshot:
    """
    Retrieve and maintain liquidity positions for Uniswap V4 pools.
    """

    def __init__(
        self,
        path: pathlib.Path | str,
        chain_id: ChainId | None = None,
    ) -> None:
        def _get_pool_map(pool_identifier: ManagedPoolIdentifier) -> LiquidityMap:
            """
            Get the liquidity map for the pool, identified by an ordered tuple of the pool manager
            address and the pool ID.
            """

            _pool_manager, _pool_id = pool_identifier

            if self._file_snapshot is not None and _pool_id in self._file_snapshot:
                return LiquidityMap(
                    tick_bitmap={
                        int(k): UniswapV4BitmapAtWord(**v)
                        for k, v in self._file_snapshot[_pool_id]["tick_bitmap"].items()
                    },
                    tick_data={
                        int(k): UniswapV4LiquidityAtTick(**v)
                        for k, v in self._file_snapshot[_pool_id]["tick_data"].items()
                    },
                )
            if self._dir_path is not None and (
                (pool_map_file := self._dir_path / f"{_pool_id}.json").exists()
            ):
                pool_liquidity_snapshot = pydantic_core.from_json(pool_map_file.read_bytes())
                return LiquidityMap(
                    tick_bitmap={
                        int(k): UniswapV4BitmapAtWord(**v)
                        for k, v in pool_liquidity_snapshot["tick_bitmap"].items()
                    },
                    tick_data={
                        int(k): UniswapV4LiquidityAtTick(**v)
                        for k, v in pool_liquidity_snapshot["tick_data"].items()
                    },
                )

            return LiquidityMap(
                tick_bitmap={},
                tick_data={},
            )

        path = pathlib.Path(path)

        self.newest_block: BlockNumber
        self._dir_path: pathlib.Path | None = None
        self._file_snapshot: dict[str, Any] | None = None

        if path.is_file():
            self._file_snapshot = pydantic_core.from_json(path.read_bytes())
            self.pool_manager = self._file_snapshot.pop("pool_manager")
            self.newest_block = self._file_snapshot.pop("snapshot_block")
        if path.is_dir():
            metadata_path = path / "_metadata.json"
            metadata = pydantic_core.from_json(metadata_path.read_bytes())
            self.pool_manager = metadata["pool_manager"]
            self.newest_block = metadata["snapshot_block"]
            self._dir_path = path.absolute()

        self._chain_id = chain_id if chain_id is not None else connection_manager.default_chain_id

        logger.info(
            f"Loaded Uniswap V4 LP snapshot: {len(self._file_snapshot) if self._file_snapshot else len(tuple(path.glob('0x*.json')))} pools @ block {self.newest_block}"  # noqa:E501
        )

        self._liquidity_events: dict[ManagedPoolIdentifier, list[UniswapV4LiquidityEvent]] = (
            defaultdict(list)
        )
        self._liquidity_snapshot: dict[ManagedPoolIdentifier, LiquidityMap] = KeyedDefaultDict(
            _get_pool_map
        )

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def pools(self) -> set[ManagedPoolIdentifier]:
        return set(self._liquidity_snapshot.keys())

    def fetch_new_events(
        self,
        to_block: BlockNumber,
        blocks_per_request: int | None = None,
    ) -> None:
        """
        Fetch liquidity events from the block following the last-known event to the target block
        using `eth_getLogs`. Blocks per request will be capped at `blocks_per_request`.
        """

        def _process_liquidity_event_log(
            event: BaseContractEvent, log: LogReceipt
        ) -> tuple[ChecksumAddress, str, UniswapV4LiquidityEvent]:
            """
            Decode an event log and convert to an address, pool ID, and a `UniswapV4LiquidityEvent`
            for processing with `UniswapV4Pool.update_liquidity_map`.
            """

            # ref: https://github.com/Uniswap/v4-core/blob/main/src/interfaces/IPoolManager.sol
            # event ModifyLiquidity(
            #     id,
            #     msg.sender,
            #     params.tickLower,
            #     params.tickUpper,
            #     params.liquidityDelta,
            #     params.salt
            # );

            decoded_event: EventData = event.process_log(log)
            pool_manager_address = get_checksum_address(decoded_event["address"])
            pool_id = HexBytes(decoded_event["args"]["id"]).to_0x_hex()
            tx_index = decoded_event["transactionIndex"]
            log_index = decoded_event["logIndex"]
            liquidity_block = decoded_event["blockNumber"]
            liquidity = decoded_event["args"]["liquidityDelta"]
            tick_lower = decoded_event["args"]["tickLower"]
            tick_upper = decoded_event["args"]["tickUpper"]

            return (
                pool_manager_address,
                pool_id,
                UniswapV4LiquidityEvent(
                    block_number=liquidity_block,
                    tx_index=tx_index,
                    log_index=log_index,
                    liquidity=liquidity,
                    tick_lower=tick_lower,
                    tick_upper=tick_upper,
                ),
            )

        logger.info(f"Updating Uniswap V4 snapshot from block {self.newest_block} to {to_block}")

        v4_pool_manager = Web3().eth.contract(abi=UNISWAP_V4_POOL_MANAGER_ABI)
        event = v4_pool_manager.events.ModifyLiquidity
        event_instance = event()
        event_abi = cast(
            "ABIEvent",
            get_abi_element(
                abi=v4_pool_manager.abi,
                abi_element_identifier=event.event_name,
            ),
        )

        event_logs = fetch_logs_retrying(
            w3=connection_manager.get_web3(self.chain_id),
            start_block=self.newest_block + 1,
            end_block=to_block,
            max_blocks_per_request=blocks_per_request,
            topic_signature=[HexBytes(event_abi_to_log_topic(event_abi))],
        )

        for event_log in tqdm.tqdm(
            event_logs,
            desc=f"Processing {event.event_name} events",
            unit="event",
            bar_format="{desc}: {percentage:3.1f}% |{bar}| {n_fmt}/{total_fmt}",
        ):
            pool_manager_address, pool_id, liquidity_event = _process_liquidity_event_log(
                event_instance, event_log
            )
            self._liquidity_events[(pool_manager_address, pool_id)].append(liquidity_event)

        self.newest_block = to_block

    def pending_updates(
        self,
        pool_manager: HexAddress | bytes,
        pool_id: HexStr | bytes,
    ) -> tuple[UniswapV4PoolLiquidityMappingUpdate, ...]:
        """
        Consume and return all pending liquidity events for this pool.
        """

        pool_key = get_checksum_address(pool_manager), HexBytes(pool_id).to_0x_hex()
        pending_events = tuple(self._liquidity_events[pool_key])
        self._liquidity_events[pool_key] = []

        return tuple(
            UniswapV4PoolLiquidityMappingUpdate(
                block_number=event.block_number,
                liquidity=event.liquidity,
                tick_lower=event.tick_lower,
                tick_upper=event.tick_upper,
            )
            for event in pending_events
        )

    def tick_bitmap(
        self,
        pool_manager: HexAddress | bytes,
        pool_id: HexStr | bytes,
    ) -> dict[int, UniswapV4BitmapAtWord]:
        """
        Consume the tick initialization bitmaps for the pool.
        """

        pool_key: ManagedPoolIdentifier = (
            get_checksum_address(pool_manager),
            HexBytes(pool_id).to_0x_hex(),
        )
        tick_bitmap = self._liquidity_snapshot[pool_key]["tick_bitmap"]
        self._liquidity_snapshot[pool_key]["tick_bitmap"] = {}

        return tick_bitmap

    def tick_data(
        self,
        pool_manager: HexAddress | bytes,
        pool_id: HexStr | bytes,
    ) -> dict[int, UniswapV4LiquidityAtTick]:
        """
        Consume the liquidity mapping for the pool.
        """

        pool_key: ManagedPoolIdentifier = (
            get_checksum_address(pool_manager),
            HexBytes(pool_id).to_0x_hex(),
        )
        tick_data = self._liquidity_snapshot[pool_key]["tick_data"]
        self._liquidity_snapshot[pool_key]["tick_data"] = {}

        return tick_data

    def update(
        self,
        pool_manager: HexAddress | bytes,
        pool_id: HexStr | bytes,
        tick_data: dict[int, UniswapV4LiquidityAtTick],
        tick_bitmap: dict[int, UniswapV4BitmapAtWord],
    ) -> None:
        """
        Update the liquidity mapping for the pool.
        """

        pool_key: ManagedPoolIdentifier = (
            get_checksum_address(pool_manager),
            HexBytes(pool_id).to_0x_hex(),
        )
        self._liquidity_snapshot[pool_key]["tick_bitmap"].update(tick_bitmap)
        self._liquidity_snapshot[pool_key]["tick_data"].update(tick_data)



from .base import DegenbotError, DegenbotTypeError, DegenbotValueError

# isort: split


from . import arbitrage, erc20, evm, liquidity_pool, manager, registry, transaction

__all__ = (
    "DegenbotError",
    "DegenbotTypeError",
    "DegenbotValueError",
    "arbitrage",
    "erc20",
    "evm",
    "liquidity_pool",
    "manager",
    "registry",
    "transaction",
)



class DegenbotError(Exception):
    """
    Base exception used as the parent class for all exceptions raised by this package.

    Calling code should catch `DegenbotError` and derived classes separately before general
    exceptions, e.g.:

    ```
    try:
        degenbot.some_function()
    except SpecificDegenbotError:
        ... # handle a specific exception
    except DegenbotError:
        ... # handle non-specific degenbot exception
    except Exception:
        ... # handle exceptions raised by 3rd party dependencies or Python built-ins
    ```

    An optional string-formatted message may be attached to the exception and retrieved by accessing
    the `.message` attribute.
    """

    message: str | None = None

    def __init__(self, message: str | None = None) -> None:
        if message:
            self.message = message
            super().__init__(message)


class DegenbotValueError(DegenbotError): ...


class DegenbotTypeError(DegenbotError): ...


class ExternalServiceError(DegenbotError):
    """
    Raised on errors resulting to some call to an external service.
    """

    def __init__(self, error: str) -> None:
        self.error = error
        super().__init__(message=f"External service error: {error}")



# TODO
# ----------------------------------------------------
# PRIORITY      TASK
# high          write state_update method
# high          create a manager for Curve pools
# medium        add liquidity modifying mode for external_update
# medium        investigate differences in get_dy_underlying vs exchange_underlying at GUSD-3Crv
# low           investigate providing overrides for live-lookup contracts

import contextlib
from collections.abc import Iterable, Sequence
from threading import Lock
from typing import TYPE_CHECKING, Any, cast
from weakref import WeakSet

import eth_abi.abi
import web3.exceptions
from eth_abi.exceptions import DecodingError, InsufficientDataBytes
from eth_typing import AnyAddress, ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.exceptions import Web3Exception
from web3.types import BlockIdentifier

from degenbot import connection_manager, get_checksum_address
from degenbot.constants import ZERO_ADDRESS
from degenbot.curve.deployments import (
    BROKEN_CURVE_V1_POOLS,
    CURVE_V1_FACTORY_ADDRESS,
    CURVE_V1_METAREGISTRY_ADDRESS,
    CURVE_V1_REGISTRY_ADDRESS,
)
from degenbot.curve.types import CurveStableswapPoolState, CurveStableSwapPoolStateUpdated
from degenbot.erc20 import Erc20Token, Erc20TokenManager
from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.evm import EVMRevertError
from degenbot.exceptions.arbitrage import NoLiquidity
from degenbot.exceptions.liquidity_pool import BrokenPool, InvalidSwapInputAmount
from degenbot.functions import encode_function_calldata, get_number_for_block_identifier, raw_call
from degenbot.logging import logger
from degenbot.registry import pool_registry
from degenbot.types.abstract import AbstractArbitrage, AbstractLiquidityPool
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.types.concrete import (
    AbstractPublisherMessage,
    BoundedCache,
    Publisher,
    PublisherMixin,
    Subscriber,
)


class CurveStableswapPool(PublisherMixin, AbstractLiquidityPool):
    type PoolState = CurveStableswapPoolState
    _state_cache: BoundedCache[BlockNumber, PoolState]

    # Constants from contract
    # ref: https://github.com/curvefi/curve-contract/blob/master/contracts/pool-templates/base/SwapTemplateBase.vy
    PRECISION_DECIMALS: int = 18
    PRECISION: int = 10**PRECISION_DECIMALS
    LENDING_PRECISION: int = PRECISION
    FEE_DENOMINATOR: int = 10**10
    A_PRECISION: int = 100
    MAX_COINS: int = 8

    D_VARIANT_GROUP_0 = frozenset(
        get_checksum_address(pool_address)
        for pool_address in (
            "0x06364f10B501e868329afBc005b3492902d6C763",
            "0x4CA9b3063Ec5866A4B82E437059D2C43d1be596F",
            "0x52EA46506B9CC5Ef470C5bf89f17Dc28bB35D85C",
            "0x7fC77b5c7614E1533320Ea6DDc2Eb61fa00A9714",
            "0x93054188d876f558f4a66B2EF1d97d16eDf0895B",
            "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7",
        )
    )
    D_VARIANT_GROUP_1 = frozenset(
        get_checksum_address(pool_address)
        for pool_address in (
            "0x45F783CCE6B7FF23B2ab2D70e416cdb7D6055f51",
            "0x79a8C46DeA5aDa233ABaFFD40F3A0A2B1e5A4F27",
            "0xA2B47E3D5c44877cca798226B7B8118F9BFb7A56",
            "0xA5407eAE9Ba41422680e2e00537571bcC53efBfD",
        )
    )
    D_VARIANT_GROUP_2 = frozenset(
        get_checksum_address(pool_address)
        for pool_address in (
            "0x0AD66FeC8dB84F8A3365ADA04aB23ce607ac6E24",
            "0x1c899dED01954d0959E034b62a728e7fEbE593b0",
            "0x3F1B0278A9ee595635B61817630cC19DE792f506",
            "0x3Fb78e61784C9c637D560eDE23Ad57CA1294c14a",
            "0x453D92C7d4263201C69aACfaf589Ed14202d83a4",
            "0x663aC72a1c3E1C4186CD3dCb184f216291F4878C",
            "0x6A274dE3e2462c7614702474D64d376729831dCa",
            "0x7C0d189E1FecB124487226dCbA3748bD758F98E4",
            "0x875DF0bA24ccD867f8217593ee27253280772A97",
            "0x99f5aCc8EC2Da2BC0771c32814EFF52b712de1E5",
            "0x9D0464996170c6B9e75eED71c68B99dDEDf279e8",
            "0xB37D6c07482Bc11cd28a1f11f1a6ad7b66Dec933",
            "0xB657B895B265C38c53FFF00166cF7F6A3C70587d",
            "0xD6Ac1CB9019137a896343Da59dDE6d097F710538",
            "0xE95E4c2dAC312F31Dc605533D5A4d0aF42579308",
            "0xf7b55C3732aD8b2c2dA7c24f30A69f55c54FB717",
        )
    )
    D_VARIANT_GROUP_3 = frozenset(
        get_checksum_address(pool_address)
        for pool_address in (
            "0xDC24316b9AE028F1497c275EB9192a3Ea0f67022",
            "0xDeBF20617708857ebe4F679508E7b7863a8A8EeE",
            "0xEB16Ae0052ed37f479f7fe63849198Df1765a733",
        )
    )
    D_VARIANT_GROUP_4 = frozenset(
        get_checksum_address(pool_address)
        for pool_address in (
            "0x1062FD8eD633c1f080754c19317cb3912810B5e5",
            "0x1C5F80b6B68A9E1Ef25926EeE00b5255791b996B",
            "0x26f3f26F46cBeE59d1F8860865e13Aa39e36A8c0",
            "0x2d600BbBcC3F1B6Cb9910A70BaB59eC9d5F81B9A",
            "0x320B564Fb9CF36933eC507a846ce230008631fd3",
            "0x3b21C2868B6028CfB38Ff86127eF22E68d16d53B",
            "0x69ACcb968B19a53790f43e57558F5E443A91aF22",
            "0x971add32Ea87f10bD192671630be3BE8A11b8623",
            "0xCA0253A98D16e9C1e3614caFDA19318EE69772D0",
            "0xfBB481A443382416357fA81F16dB5A725DC6ceC8",
        )
    )

    Y_VARIANT_GROUP_0 = frozenset(
        get_checksum_address(pool_address)
        for pool_address in (
            "0x45F783CCE6B7FF23B2ab2D70e416cdb7D6055f51",
            "0x52EA46506B9CC5Ef470C5bf89f17Dc28bB35D85C",
            "0x79a8C46DeA5aDa233ABaFFD40F3A0A2B1e5A4F27",
            "0xA2B47E3D5c44877cca798226B7B8118F9BFb7A56",
            "0xA5407eAE9Ba41422680e2e00537571bcC53efBfD",
        )
    )
    Y_VARIANT_GROUP_1 = frozenset(
        get_checksum_address(pool_address)
        for pool_address in (
            "0x06364f10B501e868329afBc005b3492902d6C763",
            "0x45F783CCE6B7FF23B2ab2D70e416cdb7D6055f51",
            "0x4CA9b3063Ec5866A4B82E437059D2C43d1be596F",
            "0x52EA46506B9CC5Ef470C5bf89f17Dc28bB35D85C",
            "0x79a8C46DeA5aDa233ABaFFD40F3A0A2B1e5A4F27",
            "0x7fC77b5c7614E1533320Ea6DDc2Eb61fa00A9714",
            "0x93054188d876f558f4a66B2EF1d97d16eDf0895B",
            "0xA2B47E3D5c44877cca798226B7B8118F9BFb7A56",
            "0xA5407eAE9Ba41422680e2e00537571bcC53efBfD",
            "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7",
        )
    )

    Y_D_VARIANT_GROUP_0 = frozenset(
        get_checksum_address(pool_address)
        for pool_address in (
            "0xDcEF968d416a41Cdac0ED8702fAC8128A64241A2",
            "0xf253f83AcA21aAbD2A20553AE0BF7F65C755A07F",
        )
    )

    def _notify_subscribers(self: Publisher, message: AbstractPublisherMessage) -> None:
        for subscriber in self._subscribers:
            subscriber.notify(publisher=self, message=message)

    def __init__(
        self,
        address: ChecksumAddress | str,
        *,
        chain_id: ChainId | None = None,
        state_block: BlockNumber | None = None,
        silent: bool = False,
        state_cache_depth: int = 8,
    ) -> None:
        """
        A Curve V1 (StableSwap) pool.

        Arguments
        ---------
        address:
            Address for the deployed pool contract.
        chain_id:
            The chain ID where the pool contract is deployed.
        state_block:
            Fetch initial state values from the chain at a particular block height. Defaults to the
            latest block if omitted.
        silent:
            Suppress status output.
        state_cache_depth:
            How many unique block-state pairs to hold in the state cache.
        """

        self._chain_id = chain_id if chain_id is not None else connection_manager.default_chain_id
        w3 = connection_manager.get_web3(self.chain_id)
        if state_block is None:
            state_block = w3.eth.block_number

        self.fee_gamma: int
        self.mid_fee: int
        self.offpeg_fee_multiplier: int
        self.out_fee: int
        self.precision_multipliers: tuple[int, ...]
        self.rate_multipliers: tuple[int, ...]
        self.use_lending: tuple[bool, ...]
        self.oracle_method: int | None

        self.initial_a_coefficient: int | None = None
        self.initial_a_coefficient_time: int | None = None
        self.future_a_coefficient: int | None = None
        self.future_a_coefficient_time: int | None = None

        def get_a_scaling_values() -> None:
            with (
                contextlib.suppress(Web3Exception, DecodingError),
                w3.batch_requests() as batch,
            ):
                batch.add(
                    w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="initial_A()",
                                function_arguments=None,
                            ),
                        },
                        block_identifier=self.update_block,
                    )
                )
                batch.add(
                    w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="initial_A_time()",
                                function_arguments=None,
                            ),
                        },
                        block_identifier=self.update_block,
                    )
                )
                batch.add(
                    w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="future_A()",
                                function_arguments=None,
                            ),
                        },
                        block_identifier=self.update_block,
                    )
                )
                batch.add(
                    w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="future_A_time()",
                                function_arguments=None,
                            ),
                        },
                        block_identifier=self.update_block,
                    )
                )

                initial_a, initial_a_time, future_a, future_a_time = batch.execute()

                (initial_a,) = eth_abi.abi.decode(
                    types=["uint256"], data=cast("HexBytes", initial_a)
                )
                (initial_a_time,) = eth_abi.abi.decode(
                    types=["uint256"], data=cast("HexBytes", initial_a_time)
                )
                (future_a,) = eth_abi.abi.decode(types=["uint256"], data=cast("HexBytes", future_a))
                (future_a_time,) = eth_abi.abi.decode(
                    types=["uint256"], data=cast("HexBytes", future_a_time)
                )

                self.initial_a_coefficient = cast("int", initial_a)
                self.initial_a_coefficient_time = cast("int", initial_a_time)
                self.future_a_coefficient = cast("int", future_a)
                self.future_a_coefficient_time = cast("int", future_a_time)

        self.a_coefficient: int
        self.fee: int
        self.admin_fee: int

        def get_coefficient_and_fees() -> None:
            with w3.batch_requests() as batch:
                batch.add(
                    w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="A()",
                                function_arguments=None,
                            ),
                        },
                        block_identifier=self.update_block,
                    )
                )
                batch.add(
                    w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="fee()",
                                function_arguments=None,
                            ),
                        },
                        block_identifier=self.update_block,
                    )
                )
                batch.add(
                    w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="admin_fee()",
                                function_arguments=None,
                            ),
                        },
                        block_identifier=self.update_block,
                    )
                )

                a_coefficient, pool_fee, admin_fee = batch.execute()

                (a_coefficient,) = eth_abi.abi.decode(
                    types=["uint256"], data=cast("HexBytes", a_coefficient)
                )
                (pool_fee,) = eth_abi.abi.decode(types=["uint256"], data=cast("HexBytes", pool_fee))
                (admin_fee,) = eth_abi.abi.decode(
                    types=["uint256"], data=cast("HexBytes", admin_fee)
                )

                self.a_coefficient = cast("int", a_coefficient)
                self.fee = cast("int", pool_fee)
                self.admin_fee = cast("int", admin_fee)

        def get_coin_index_type() -> str:
            # Identify the coins input format (int128 or uint256)
            # Some contracts accept token_id as an int128, some accept uint256

            _type = "uint256"
            with contextlib.suppress(InsufficientDataBytes, web3.exceptions.ContractLogicError):
                eth_abi.abi.decode(
                    types=["address"],
                    data=w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": Web3.keccak(text=f"coins({_type})")[:4]
                            + eth_abi.abi.encode(types=[_type], args=[0]),
                        },
                        block_identifier=state_block,
                    ),
                )

                return _type

            _type = "int128"
            with contextlib.suppress(InsufficientDataBytes, web3.exceptions.ContractLogicError):
                eth_abi.abi.decode(
                    types=["address"],
                    data=w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": Web3.keccak(text=f"coins({_type})")[:4]
                            + eth_abi.abi.encode(types=[_type], args=[0]),
                        },
                        block_identifier=state_block,
                    ),
                )

                return _type

            raise DegenbotValueError(
                message="Could not determine input type for pool"
            )  # pragma: no cover

        def get_token_addresses() -> tuple[ChecksumAddress, ...]:
            token_addresses = []
            for token_id in range(self.MAX_COINS):  # pragma: no branch
                try:
                    token_address: str
                    (token_address,) = raw_call(
                        w3=w3,
                        address=self.address,
                        calldata=encode_function_calldata(
                            function_prototype=f"coins({self._coin_index_type})",
                            function_arguments=[token_id],
                        ),
                        return_types=["address"],
                        block_identifier=state_block,
                    )
                    token_addresses.append(get_checksum_address(token_address))
                except web3.exceptions.ContractLogicError:
                    break

            return tuple(token_addresses)

        def get_lp_token_address() -> ChecksumAddress:
            for contract_address in (
                CURVE_V1_METAREGISTRY_ADDRESS,
                CURVE_V1_REGISTRY_ADDRESS,
                CURVE_V1_FACTORY_ADDRESS,
            ):
                with contextlib.suppress(Web3Exception, DecodingError):
                    (lp_token_address,) = raw_call(
                        w3=connection_manager.get_web3(chain_id=self.chain_id),
                        address=contract_address,
                        calldata=encode_function_calldata(
                            function_prototype="get_lp_token(address)",
                            function_arguments=[self.address],
                        ),
                        return_types=["address"],
                        block_identifier=state_block,
                    )
                    if lp_token_address == ZERO_ADDRESS:
                        continue
                    return get_checksum_address(lp_token_address)

            raise DegenbotValueError(
                message=f"Could not identify LP token for pool {self.address}"
            )  # pragma: no cover

        def get_pool_from_lp_token(token: AnyAddress) -> ChecksumAddress:
            for contract_address in (
                CURVE_V1_METAREGISTRY_ADDRESS,
                CURVE_V1_REGISTRY_ADDRESS,
                CURVE_V1_FACTORY_ADDRESS,
            ):
                with contextlib.suppress(Web3Exception, DecodingError):
                    (pool_address,) = raw_call(
                        w3=connection_manager.get_web3(chain_id=self.chain_id),
                        address=contract_address,
                        calldata=encode_function_calldata(
                            function_prototype="get_pool_from_lp_token(address)",
                            function_arguments=[get_checksum_address(token)],
                        ),
                        return_types=["address"],
                        block_identifier=state_block,
                    )
                    return get_checksum_address(pool_address)

            raise DegenbotValueError(
                message=f"Could not identify base pool from LP token for {self.address}"
            )  # pragma: no cover

        def is_metapool() -> bool:
            """
            Check if the registry contract and the factory contract report that this is registered
            as a metapool. Some metapools are not correctly marked in one of the contracts, so this
            function checks both.
            """
            w3 = connection_manager.get_web3(chain_id=self.chain_id)

            is_meta_results = [False]

            for contract_address in (
                CURVE_V1_METAREGISTRY_ADDRESS,
                CURVE_V1_REGISTRY_ADDRESS,
                CURVE_V1_FACTORY_ADDRESS,
            ):
                with contextlib.suppress(Web3Exception, DecodingError):
                    (result,) = raw_call(
                        w3=w3,
                        address=contract_address,
                        calldata=encode_function_calldata(
                            function_prototype="is_meta(address)",
                            function_arguments=[self.address],
                        ),
                        return_types=["bool"],
                        block_identifier=state_block,
                    )
                    is_meta_results.append(cast("bool", result))

            return any(is_meta_results)

        def set_pool_specific_attributes() -> None:
            match self.address:
                case "0xA2B47E3D5c44877cca798226B7B8118F9BFb7A56":
                    self.use_lending = (True, True)
                    self.precision_multipliers = (1, 10**12)
                case "0x80466c64868E1ab14a1Ddf27A676C3fcBE638Fe5":
                    self.fee_gamma = 10000000000000000
                    self.mid_fee = 4000000
                    self.out_fee = 40000000
                case "0xDcEF968d416a41Cdac0ED8702fAC8128A64241A2":
                    self.precision_multipliers = (1, 1000000000000)
                case "0x52EA46506B9CC5Ef470C5bf89f17Dc28bB35D85C":
                    self.use_lending = (True, True, False)
                    self.precision_multipliers = (1, 10**12, 10**12)
                case "0x06364f10B501e868329afBc005b3492902d6C763":
                    self.use_lending = (True, True, True, False)
                case "0xDeBF20617708857ebe4F679508E7b7863a8A8EeE":
                    self.precision_multipliers = (1, 10**12, 10**12)
                    (self.offpeg_fee_multiplier,) = eth_abi.abi.decode(
                        types=["uint256"],
                        data=w3.eth.call(
                            transaction={
                                "to": self.address,
                                "data": Web3.keccak(text="offpeg_fee_multiplier()")[:4],
                            },
                            block_identifier=state_block,
                        ),
                    )
                case "0x2dded6Da1BF5DBdF597C45fcFaa3194e53EcfeAF":
                    self.precision_multipliers = (1, 10**12, 10**12)
                case "0x79a8C46DeA5aDa233ABaFFD40F3A0A2B1e5A4F27":
                    self.precision_multipliers = (1, 10**12, 10**12, 1)
                    self.use_lending = (True, True, True, True)
                case "0x45F783CCE6B7FF23B2ab2D70e416cdb7D6055f51":
                    self.precision_multipliers = (1, 10**12, 10**12, 1)
                    self.use_lending = (True, True, True, True)
                case "0xA5407eAE9Ba41422680e2e00537571bcC53efBfD":
                    self.use_lending = (False, False, False, False)
                case (
                    "0x59Ab5a5b5d617E478a2479B0cAD80DA7e2831492"
                    | "0xBfAb6FA95E0091ed66058ad493189D2cB29385E6"
                ):
                    self._set_oracle_method(block_number=state_block)
                case "0xEB16Ae0052ed37f479f7fe63849198Df1765a733":
                    (self.offpeg_fee_multiplier,) = eth_abi.abi.decode(
                        types=["uint256"],
                        data=w3.eth.call(
                            transaction={
                                "to": self.address,
                                "data": Web3.keccak(text="offpeg_fee_multiplier()")[:4],
                            },
                            block_identifier=state_block,
                        ),
                    )

        self.address = get_checksum_address(address)
        if self.address in BROKEN_CURVE_V1_POOLS:
            raise BrokenPool

        _block = w3.eth.get_block(state_block)

        self._create_timestamp = _block["timestamp"]

        self._block_timestamps: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_admin_balances: BoundedCache[BlockNumber, tuple[int, ...]] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_base_cache_updated: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_base_virtual_price: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_contract_D: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_gamma: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_price_scale: BoundedCache[BlockNumber, tuple[int, ...]] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_rates_from_aeth: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_rates_from_ctokens: BoundedCache[BlockNumber, tuple[int, ...]] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_rates_from_cytokens: BoundedCache[BlockNumber, tuple[int, ...]] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_rates_from_oracle: BoundedCache[BlockNumber, tuple[int, ...]] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_rates_from_reth: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_rates_from_ytokens: BoundedCache[BlockNumber, tuple[int, ...]] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_scaled_redemption_price: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth
        )
        self._cached_virtual_price: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth
        )

        # token setup
        self._coin_index_type = get_coin_index_type()

        token_manager = Erc20TokenManager(chain_id=self.chain_id)
        self.lp_token = token_manager.get_erc20token(
            address=get_lp_token_address(),
            silent=silent,
        )
        self.tokens: tuple[Erc20Token, ...] = tuple(
            token_manager.get_erc20token(
                address=token_address,
                silent=silent,
            )
            for token_address in get_token_addresses()
        )

        # metapool setup
        self.base_pool: CurveStableswapPool | None = None
        if is_metapool():
            # Curve metapools hold the LP token for the base pool at index 1
            base_pool_address = get_pool_from_lp_token(self.tokens[1].address)

            if (
                base_pool := pool_registry.get(
                    pool_address=base_pool_address, chain_id=self.chain_id
                )
            ) is None:
                base_pool = CurveStableswapPool(
                    base_pool_address, state_block=state_block, silent=silent
                )
            if TYPE_CHECKING:
                assert isinstance(base_pool, CurveStableswapPool)

            self.base_pool = base_pool
            self.tokens_underlying = (self.tokens[0], *self.base_pool.tokens)

            self.base_cache_updated: int | None = None
            with contextlib.suppress(web3.exceptions.ContractLogicError):
                self.base_cache_updated = self._get_base_cache_updated(block_number=state_block)

            self.base_virtual_price: int
            with contextlib.suppress(web3.exceptions.ContractLogicError):
                self.base_virtual_price = self._get_base_virtual_price(block_number=state_block)

        _balances = []
        for token_id, _ in enumerate(self.tokens):
            (token_balance,) = eth_abi.abi.decode(
                types=[self._coin_index_type],
                data=w3.eth.call(
                    transaction={
                        "to": self.address,
                        "data": Web3.keccak(text=f"balances({self._coin_index_type})")[:4]
                        + eth_abi.abi.encode(types=[self._coin_index_type], args=[token_id]),
                    },
                    block_identifier=state_block,
                ),
            )
            _balances.append(token_balance)

        """
        3pool example
        rate_multipliers = [
          10**12000000,             <------ 10**18 == 10**(18 + 18 - 18)
          10**12000000000000000000, <------ 10**30 == 10**(18 + 18 - 6)
          10**12000000000000000000, <------ 10**30 == 10**(18 + 18 - 6)
        ]
        """
        self.rate_multipliers = tuple(
            10 ** (2 * self.PRECISION_DECIMALS - token.decimals) for token in self.tokens
        )
        self.precision_multipliers = tuple(
            cast("int", 10 ** (self.PRECISION_DECIMALS - token.decimals)) for token in self.tokens
        )

        self._state = CurveStableswapPoolState(
            address=self.address,
            balances=tuple(_balances),
            block=state_block,
        )
        self._state_cache = BoundedCache(max_items=state_cache_depth)
        self._state_cache[state_block] = self._state
        self._state_lock = Lock()
        self._block_timestamps[state_block] = _block["timestamp"]

        get_a_scaling_values()
        get_coefficient_and_fees()
        set_pool_specific_attributes()

        fee_string = f"{100 * self.fee / self.FEE_DENOMINATOR:.2f}"
        token_string = "-".join([token.symbol for token in self.tokens])
        self.name = f"{token_string} ({self.__class__.__name__}, {fee_string}%)"

        self._subscribers: WeakSet[Subscriber] = WeakSet()

        pool_registry.add(pool_address=self.address, chain_id=self.chain_id, pool=self)

        if not silent:
            logger.info(
                f"{self.name} @ {self.address}, A={self.a_coefficient}, fee={100 * self.fee / self.FEE_DENOMINATOR:.2f}%"  # noqa:E501
            )
            for token_id, (token, balance) in enumerate(
                zip(self.tokens, self.balances, strict=True)
            ):
                logger.info(f"• Token {token_id}: {token} - Reserves: {balance}")

    def __getstate__(self) -> dict[str, Any]:
        # Remove objects that cannot be pickled and are unnecessary to perform
        # the calculation
        dropped_attributes = (
            "_state_lock",
            "_subscribers",
        )

        with self._state_lock:
            return {k: v for k, v in self.__dict__.items() if k not in dropped_attributes}

    def __repr__(self) -> str:  # pragma: no cover
        token_string = "-".join([token.symbol for token in self.tokens])
        return f"{self.__class__.__name__}(address={self.address}, tokens={token_string}, fee={100 * self.fee / self.FEE_DENOMINATOR:.2f}%, A={self.a_coefficient})"  # noqa:E501

    @property
    def balances(self) -> tuple[int, ...]:
        return self.state.balances

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def state(self) -> CurveStableswapPoolState:
        return self._state

    @property
    def update_block(self) -> BlockNumber:
        if TYPE_CHECKING:
            assert self.state.block is not None
        return self.state.block

    def _a(self, timestamp: int | None = None) -> int:
        """
        Handle ramping A up or down
        """

        if any(
            [
                self.future_a_coefficient is None,
                self.initial_a_coefficient is None,
            ]
        ):
            return self.a_coefficient * self.A_PRECISION

        if TYPE_CHECKING:
            assert self.initial_a_coefficient is not None
            assert self.initial_a_coefficient_time is not None
            assert self.future_a_coefficient_time is not None
            assert self.future_a_coefficient is not None

        if self._create_timestamp >= self.future_a_coefficient_time:
            return self.future_a_coefficient

        if timestamp is None:
            timestamp = connection_manager.get_web3(self.chain_id).eth.get_block("latest")[
                "timestamp"
            ]

        a_1 = self.future_a_coefficient
        t_1 = self.future_a_coefficient_time

        # Modified from contract template to check timestamp argument instead
        # of block.timestamp
        if timestamp < t_1:
            a_0 = self.initial_a_coefficient
            t_0 = self.initial_a_coefficient_time
            if a_1 > a_0:
                scaled_a = a_0 + (a_1 - a_0) * (timestamp - t_0) // (t_1 - t_0)
            else:
                scaled_a = a_0 - (a_0 - a_1) * (timestamp - t_0) // (t_1 - t_0)
        else:
            scaled_a = a_1

        return scaled_a

    def calc_token_amount(
        self,
        *,
        amounts: Sequence[int],
        deposit: bool,
        block_identifier: BlockIdentifier | None = None,
        override_state: CurveStableswapPoolState | None = None,
    ) -> int:
        """
        Simplified method to calculate addition or reduction in token supply at
        deposit or withdrawal without taking fees into account (but looking at
        slippage).
        Needed to prevent front-running, not for precise calculations!
        """

        n_coins = len(self.tokens)

        pool_balances = (
            list(override_state.balances) if override_state is not None else list(self.balances)
        )

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else get_number_for_block_identifier(
                block_identifier,
                connection_manager.get_web3(self.chain_id),
            )
        )

        xp = self._xp(rates=self.rate_multipliers, balances=pool_balances)
        amp = self._a(timestamp=self._block_timestamps[block_number])
        d_0 = self._get_d(_xp=xp, _amp=amp)

        for i in range(n_coins):
            if deposit:
                pool_balances[i] += amounts[i]
            else:
                pool_balances[i] -= amounts[i]

        xp = self._xp(rates=self.rate_multipliers, balances=pool_balances)
        d_1 = self._get_d(xp, amp)
        token_amount: int = self.lp_token.get_total_supply(block_identifier=block_number)

        diff = d_1 - d_0 if deposit else d_0 - d_1

        return diff * token_amount // d_0

    def calc_withdraw_one_coin(
        self, _token_amount: int, i: int, block_identifier: BlockIdentifier | None = None
    ) -> tuple[int, ...]:
        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else get_number_for_block_identifier(
                block_identifier,
                connection_manager.get_web3(self.chain_id),
            )
        )

        n_coins = len(self.tokens)
        amp = self._a(timestamp=self._block_timestamps[block_number])
        total_supply = self.lp_token.get_total_supply(block_identifier=block_number)
        precisions = self.precision_multipliers
        xp = self._xp(rates=self.rate_multipliers, balances=self.balances)
        d_0 = self._get_d(xp, amp)
        d_1 = d_0 - _token_amount * d_0 // total_supply
        new_y = self._get_y_d(amp, i, xp, d_1)
        dy_0 = (xp[i] - new_y) // precisions[i]

        xp_reduced = list(xp)
        _fee = self.fee * n_coins // (4 * (n_coins - 1))
        for j in range(n_coins):
            dx_expected = xp[j] * d_1 // d_0 - new_y if j == i else xp[j] - xp[j] * d_1 // d_0
            xp_reduced[j] -= _fee * dx_expected // self.FEE_DENOMINATOR

        dy = xp_reduced[i] - self._get_y_d(amp, i, xp_reduced, d_1)
        dy = (dy - 1) // precisions[i]

        return dy, dy_0 - dy, total_supply

    def _get_scaled_redemption_price(self, block_number: BlockNumber) -> int:
        with contextlib.suppress(KeyError):
            return self._cached_scaled_redemption_price[block_number]

        redemption_price_scale = 10**9

        w3 = connection_manager.get_web3(self.chain_id)

        snap_contract_address: str
        (snap_contract_address,) = eth_abi.abi.decode(
            types=["address"],
            data=w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="redemption_price_snap()")[:4],
                },
                block_identifier=block_number,
            ),
        )

        rate: int
        (rate,) = eth_abi.abi.decode(
            types=["uint256"],
            data=w3.eth.call(
                transaction={
                    "to": get_checksum_address(snap_contract_address),
                    "data": Web3.keccak(text="snappedRedemptionPrice()")[:4],
                },
                block_identifier=block_number,
            ),
        )
        result = rate // redemption_price_scale
        self._cached_scaled_redemption_price[block_number] = result
        return result

    def get_dy(
        self,
        i: int,
        j: int,
        dx: int,
        block_identifier: BlockIdentifier | None = None,
        override_state: CurveStableswapPoolState | None = None,
    ) -> int:
        """
        @notice Calculate the current output dy given input dx
        @dev Index values can be found via the `coins` public getter method
        @param i Index value for the coin to send
        @param j Index value of the coin to recieve
        @param dx Amount of `i` being exchanged
        @return Amount of `j` predicted

        Reference: https://github.com/curveresearch/notes/blob/main/stableswap.pdf
        """

        def _dynamic_fee(xpi: int, xpj: int, _fee: int, _feemul: int) -> int:
            if _feemul <= self.FEE_DENOMINATOR:
                return _fee
            xps2 = (xpi + xpj) ** 2
            return (_feemul * _fee) // (
                (_feemul - self.FEE_DENOMINATOR) * 4 * xpi * xpj // xps2 + self.FEE_DENOMINATOR
            )

        pool_balances = override_state.balances if override_state is not None else self.balances
        rates = self.rate_multipliers

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else get_number_for_block_identifier(
                block_identifier,
                connection_manager.get_web3(self.chain_id),
            )
        )

        if self.base_pool is not None:
            if self.address in ("0xC61557C5d177bd7DC889A3b621eEC333e168f68A",):
                rates = (
                    self.PRECISION,
                    self._get_virtual_price(block_number=block_number),
                )
            elif self.address in ("0x618788357D0EBd8A37e763ADab3bc575D54c2C7d",):
                rates = (
                    self._get_scaled_redemption_price(block_number=block_number),
                    self._get_virtual_price(block_number=block_number),
                )
            else:
                rates = (
                    self.rate_multipliers[0],
                    self._get_virtual_price(block_number=block_number),
                )

            xp = self._xp(rates=rates, balances=pool_balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = xp[j] - y - 1
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return (dy - fee) * self.PRECISION // rates[j]

        if self.address in (
            "0x4e0915C88bC70750D68C481540F081fEFaF22273",
            "0x1005F7406f32a61BD760CfA14aCCd2737913d546",
            "0x6A274dE3e2462c7614702474D64d376729831dCa",
            "0xb9446c4Ef5EBE66268dA6700D26f96273DE3d571",
            "0x3Fb78e61784C9c637D560eDE23Ad57CA1294c14a",
        ):
            live_balances = [
                token.get_balance(self.address, block_identifier=block_number)
                for token in self.tokens
            ]
            admin_balances = self._get_admin_balances(block_number=block_number)

            balances = [
                pool_balance - admin_balance
                for pool_balance, admin_balance in zip(live_balances, admin_balances, strict=True)
            ]

            xp = self._xp(rates=rates, balances=balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = xp[j] - y - 1
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return (dy - fee) * self.PRECISION // rates[j]

        if self.address == "0x80466c64868E1ab14a1Ddf27A676C3fcBE638Fe5":
            # TODO: check if any functions (price_scale, gamma, D, fee_calc) can be calculated
            # off-chain

            def _d(block_number: BlockNumber) -> int:
                with contextlib.suppress(KeyError):
                    return self._cached_contract_D[block_number]

                w3 = connection_manager.get_web3(self.chain_id)

                d: int
                (d,) = eth_abi.abi.decode(
                    types=["uint256"],
                    data=w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": Web3.keccak(text="D()")[:4],
                        },
                        block_identifier=block_number,
                    ),
                )
                self._cached_contract_D[block_number] = d
                return d

            def _gamma(block_number: BlockNumber) -> int:
                with contextlib.suppress(KeyError):
                    return self._cached_gamma[block_number]

                w3 = connection_manager.get_web3(self.chain_id)

                gamma: int
                (gamma,) = eth_abi.abi.decode(
                    types=["uint256"],
                    data=w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": Web3.keccak(text="gamma()")[:4],
                        },
                        block_identifier=block_number,
                    ),
                )
                self._cached_gamma[block_number] = gamma
                return gamma

            def _price_scale(block_number: BlockNumber) -> tuple[int, ...]:
                with contextlib.suppress(KeyError):
                    return self._cached_price_scale[block_number]

                n_coins = len(self.tokens)

                w3 = connection_manager.get_web3(self.chain_id)

                price_scale = [0] * (n_coins - 1)
                for token_index in range(n_coins - 1):
                    (price_scale[token_index],) = eth_abi.abi.decode(
                        types=["uint256"],
                        data=w3.eth.call(
                            transaction={
                                "to": self.address,
                                "data": Web3.keccak(text="price_scale(uint256)")[:4]
                                + eth_abi.abi.encode(
                                    types=["uint256"],
                                    args=[token_index],
                                ),
                            },
                            block_identifier=block_number,
                        ),
                    )
                self._cached_price_scale[block_number] = tuple(price_scale)
                return tuple(price_scale)

            def _newton_y(ann: int, gamma: int, xp: Sequence[int], d: int, token_index: int) -> int:
                """
                Calculating xp[i] given other balances xp[0..N_COINS-1] and invariant D
                _ann = A * N**N
                """

                n_coins = len(self.tokens)
                a_multiplier = self.A_PRECISION

                # Safety checks
                assert (
                    n_coins**n_coins * a_multiplier - 1
                    < ann
                    < 10000 * n_coins**n_coins * a_multiplier + 1
                ), "unsafe value for A"
                assert 10**10 - 1 < gamma < 10**16 + 1, "unsafe values for gamma"
                assert 10**17 - 1 < d < 10**15 * 10**18 + 1, "unsafe values for D"

                for index in range(3):
                    if index != token_index:
                        frac = xp[index] * 10**18 // d
                        assert 10**16 - 1 < frac < 10**20 + 1, (
                            f"{frac=} out of range"
                        )  # dev: unsafe values x[i]

                y = d // n_coins
                k_0_i = 10**18
                s_i = 0

                x_sorted = list(xp)
                x_sorted[token_index] = 0
                x_sorted = sorted(x_sorted, reverse=True)  # From high to low

                convergence_limit = max(x_sorted[0] // 10**14, d // 10**14, 100)
                for _j in range(2, n_coins + 1):
                    _x = x_sorted[n_coins - _j]
                    y = y * d // (_x * n_coins)  # Small _x first
                    s_i += _x

                for _k in range(n_coins - 1):
                    k_0_i = k_0_i * x_sorted[_k] * n_coins // d  # Large _x first

                for _ in range(255):  # pragma: no branch
                    y_prev = y

                    k_0 = k_0_i * y * n_coins // d
                    s = s_i + y

                    _g1k0 = gamma + 10**18
                    _g1k0 = _g1k0 - k_0 + 1 if _g1k0 > k_0 else k_0 - _g1k0 + 1

                    mul1 = 10**18 * d // gamma * _g1k0 // gamma * _g1k0 * a_multiplier // ann
                    mul2 = 10**18 + (2 * 10**18) * k_0 // _g1k0

                    yfprime = 10**18 * y + s * mul2 + mul1
                    _dyfprime = d * mul2

                    if yfprime < _dyfprime:
                        y = y_prev // 2
                        continue

                    yfprime -= _dyfprime
                    fprime = yfprime // y

                    y_minus = mul1 // fprime
                    y_plus = (yfprime + 10**18 * d) // fprime + y_minus * 10**18 // k_0
                    y_minus += 10**18 * s // fprime

                    y = y_prev // 2 if y_plus < y_minus else y_plus - y_minus
                    diff = y - y_prev if y > y_prev else y_prev - y

                    if diff < max(convergence_limit, y // 10**14):
                        frac = y * 10**18 // d
                        assert 10**16 - 1 < frac < 10**20 + 1, "unsafe value for y"
                        return y

                raise EVMRevertError(
                    error=f"_newton_y() did not converge for pool {self.address}"
                )  # pragma: no cover

            def _reduction_coefficient(x: Sequence[int], fee_gamma: int) -> int:
                """
                fee_gamma / (fee_gamma + (1 - K))
                where
                K = prod(x) / (sum(x) / N)**N
                (all normalized to 1e18)
                """
                k = 10**18
                s = 0
                for x_i in x:
                    s += x_i
                # Could be good to pre-sort x, but it is used only for dynamic fee,
                # so that is not so important
                for x_i in x:
                    k = k * n_coins * x_i // s
                if fee_gamma > 0:
                    k = fee_gamma * 10**18 // (fee_gamma + 10**18 - k)
                return k

            n_coins = len(self.tokens)

            assert i != j, "coin index out of range"
            assert i < n_coins, "coin index out of range"
            assert j < n_coins, "coin index out of range"
            assert dx > 0, "do not exchange 0 coins"

            precisions = [
                10**12,  # USDT
                10**10,  # WBTC
                1,  # WETH
            ]

            price_scale = _price_scale(block_number=block_number)

            _xp = list(pool_balances)
            _xp[i] += dx
            _xp[0] *= precisions[0]

            for k in range(n_coins - 1):
                _xp[k + 1] = _xp[k + 1] * price_scale[k] * precisions[k + 1] // self.PRECISION

            amp = self._a(timestamp=self._block_timestamps[block_number])
            gamma = _gamma(block_number=block_number)
            d = _d(block_number=block_number)
            y = _newton_y(amp, gamma, _xp, d, j)
            dy = _xp[j] - y - 1

            _xp[j] = y
            if j > 0:
                dy = dy * self.PRECISION // price_scale[j - 1]
            dy //= precisions[j]

            f = _reduction_coefficient(_xp, self.fee_gamma)
            fee_calc = (self.mid_fee * f + self.out_fee * (10**18 - f)) // 10**18

            dy -= fee_calc * dy // 10**10
            return dy

        if self.address in (
            "0x4CA9b3063Ec5866A4B82E437059D2C43d1be596F",
            "0x7fC77b5c7614E1533320Ea6DDc2Eb61fa00A9714",
            "0x93054188d876f558f4a66B2EF1d97d16eDf0895B",
            "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7",
        ):
            xp = self._xp(rates=rates, balances=pool_balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = (xp[j] - y - 1) * self.PRECISION // rates[j]
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return dy - fee

        if self.address in (
            "0x0Ce6a5fF5217e38315f87032CF90686C96627CAA",
            "0x19b080FE1ffA0553469D20Ca36219F17Fcf03859",
            "0x1C5F80b6B68A9E1Ef25926EeE00b5255791b996B",
            "0x1F6bb2a7a2A84d08bb821B89E38cA651175aeDd4",
            "0x21B45B2c1C53fDFe378Ed1955E8Cc29aE8cE0132",
            "0x3CFAa1596777CAD9f5004F9a0c443d912E262243",
            "0x3F1B0278A9ee595635B61817630cC19DE792f506",
            "0x4424b4A37ba0088D8a718b8fc2aB7952C7e695F5",
            "0x602a9Abb10582768Fd8a9f13aD6316Ac2A5A2e2B",
            "0x8461A004b50d321CB22B7d034969cE6803911899",
            "0x857110B5f8eFD66CC3762abb935315630AC770B5",
            "0x8818a9bb44Fbf33502bE7c15c500d0C783B73067",
            "0x9c2C8910F113181783c249d8F6Aa41b51Cde0f0c",
            "0xa1F8A6807c402E4A15ef4EBa36528A3FED24E577",
            "0xaE34574AC03A15cd58A92DC79De7B1A0800F1CE3",
            "0xAf25fFe6bA5A8a29665adCfA6D30C5Ae56CA0Cd3",
            "0xBa3436Fd341F2C8A928452Db3C5A3670d1d5Cc73",
            "0xbB2dC673E1091abCA3eaDB622b18f6D4634b2CD9",
            "0xc5424B857f758E906013F3555Dad202e4bdB4567",
            "0xc8a7C1c4B748970F57cA59326BcD49F5c9dc43E3",
            "0xcbD5cC53C5b846671C6434Ab301AD4d210c21184",
            "0xD6Ac1CB9019137a896343Da59dDE6d097F710538",
            "0xD7C10449A6D134A9ed37e2922F8474EAc6E5c100",
            "0xDC24316b9AE028F1497c275EB9192a3Ea0f67022",
            "0xDcEF968d416a41Cdac0ED8702fAC8128A64241A2",
            "0xe7A3b38c39F97E977723bd1239C3470702568e7B",
            "0xf083FBa98dED0f9C970e5a418500bad08D8b9732",
            "0xF178C0b5Bb7e7aBF4e12A4838C7b7c5bA2C623c0",
            "0xf253f83AcA21aAbD2A20553AE0BF7F65C755A07F",
            "0xfC8c34a3B3CFE1F1Dd6DBCCEC4BC5d3103b80FF0",
            "0xFD5dB7463a3aB53fD211b4af195c5BCCC1A03890",
        ):
            xp = self._xp(rates=rates, balances=pool_balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = xp[j] - y - 1
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return (dy - fee) * self.PRECISION // rates[j]

        if self.address in (
            "0x04c90C198b2eFF55716079bc06d7CCc4aa4d7512",
            "0x320B564Fb9CF36933eC507a846ce230008631fd3",
            "0x48fF31bBbD8Ab553Ebe7cBD84e1eA3dBa8f54957",
            "0x55A8a39bc9694714E2874c1ce77aa1E599461E18",
            "0x875DF0bA24ccD867f8217593ee27253280772A97",
            "0x9D0464996170c6B9e75eED71c68B99dDEDf279e8",
            "0xBaaa1F5DbA42C3389bDbc2c9D2dE134F5cD0Dc89",
            "0xDa5B670CcD418a187a3066674A8002Adc9356Ad1",
            "0xf03bD3cfE85f00bF5819AC20f0870cE8a8d1F0D8",
            "0xFB9a265b5a1f52d97838Ec7274A0b1442efAcC87",
        ):
            xp = tuple(pool_balances)
            x = xp[i] + dx
            y = self._get_y(i, j, x, xp)
            dy = xp[j] - y - 1
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return dy - fee

        if self.address in (
            "0x59Ab5a5b5d617E478a2479B0cAD80DA7e2831492",
            "0xBfAb6FA95E0091ed66058ad493189D2cB29385E6",
        ):
            live_balances = [
                token.get_balance(self.address, block_identifier=block_number)
                for token in self.tokens
            ]
            admin_balances = self._get_admin_balances(block_number=block_number)
            balances = [
                pool_balance - admin_balance
                for pool_balance, admin_balance in zip(live_balances, admin_balances, strict=True)
            ]
            rates = self._stored_rates_from_oracle(block_number=block_number)
            xp = self._xp(rates=rates, balances=balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = xp[j] - y - 1
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return (dy - fee) * self.PRECISION // rates[j]

        if self.address in (
            "0x52EA46506B9CC5Ef470C5bf89f17Dc28bB35D85C",
            "0xA2B47E3D5c44877cca798226B7B8118F9BFb7A56",
            "0xA5407eAE9Ba41422680e2e00537571bcC53efBfD",
        ):
            rates = self._stored_rates_from_ctokens(block_number=block_number)
            xp = self._xp(rates=rates, balances=pool_balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = (xp[j] - y) * self.PRECISION // rates[j]
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return dy - fee

        if self.address in ("0x2dded6Da1BF5DBdF597C45fcFaa3194e53EcfeAF",):
            rates = self._stored_rates_from_cytokens(block_number=block_number)
            xp = self._xp(rates=rates, balances=pool_balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = xp[j] - y - 1
            return (dy - (self.fee * dy // self.FEE_DENOMINATOR)) * self.PRECISION // rates[j]

        if self.address in ("0x06364f10B501e868329afBc005b3492902d6C763",):
            rates = self._stored_rates_from_ytokens(block_number=block_number)
            xp = self._xp(rates=rates, balances=pool_balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = (xp[j] - y - 1) * self.PRECISION // rates[j]
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return dy - fee

        if self.address in (
            "0x45F783CCE6B7FF23B2ab2D70e416cdb7D6055f51",
            "0x79a8C46DeA5aDa233ABaFFD40F3A0A2B1e5A4F27",
        ):
            rates = self._stored_rates_from_ytokens(block_number=block_number)
            xp = self._xp(rates=rates, balances=pool_balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = (xp[j] - y) * self.PRECISION // rates[j]
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return dy - fee

        if self.address in ("0xA96A65c051bF88B4095Ee1f2451C2A9d43F53Ae2",):
            rates = self._stored_rates_from_aeth(block_number=block_number)
            xp = self._xp(rates=rates, balances=pool_balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = xp[j] - y
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return (dy - fee) * self.PRECISION // rates[j]

        if self.address in ("0xF9440930043eb3997fc70e1339dBb11F341de7A8",):
            rates = self._stored_rates_from_reth(block_number=block_number)
            xp = self._xp(rates=rates, balances=pool_balances)
            x = xp[i] + (dx * rates[i] // self.PRECISION)
            y = self._get_y(i, j, x, xp)
            dy = xp[j] - y
            fee = self.fee * dy // self.FEE_DENOMINATOR
            return (dy - fee) * self.PRECISION // rates[j]

        if self.address in ("0xEB16Ae0052ed37f479f7fe63849198Df1765a733",):
            live_balances = [
                token.get_balance(self.address, block_identifier=block_number)
                for token in self.tokens
            ]
            admin_balances = self._get_admin_balances(block_number=block_number)

            _xp = [
                pool_balance - admin_balance
                for pool_balance, admin_balance in zip(live_balances, admin_balances, strict=True)
            ]
            x = _xp[i] + dx
            y = self._get_y(i, j, x, _xp)
            dy = _xp[j] - y
            _fee = (
                _dynamic_fee(
                    xpi=(_xp[i] + x) // 2,
                    xpj=(_xp[j] + y) // 2,
                    _fee=self.fee,
                    _feemul=self.offpeg_fee_multiplier,
                )
                * dy
                // self.FEE_DENOMINATOR
            )
            return dy - _fee

        if self.address in ("0xDeBF20617708857ebe4F679508E7b7863a8A8EeE",):
            live_balances = [
                token.get_balance(self.address, block_identifier=block_number)
                for token in self.tokens
            ]
            admin_balances = self._get_admin_balances(block_number=block_number)
            balances = [
                pool_balance - admin_balance
                for pool_balance, admin_balance in zip(live_balances, admin_balances, strict=True)
            ]

            _xp = [
                balance * rate
                for balance, rate in zip(balances, self.precision_multipliers, strict=True)
            ]

            x = _xp[i] + dx * self.precision_multipliers[i]
            y = self._get_y(i, j, x, _xp)
            dy = (_xp[j] - y) // self.precision_multipliers[j]

            _fee = (
                _dynamic_fee(
                    xpi=(_xp[i] + x) // 2,
                    xpj=(_xp[j] + y) // 2,
                    _fee=self.fee,
                    _feemul=self.offpeg_fee_multiplier,
                )
                * dy
                // self.FEE_DENOMINATOR
            )
            return dy - _fee

        # default pool behavior
        xp = self._xp(rates=rates, balances=pool_balances)
        x = xp[i] + (dx * rates[i] // self.PRECISION)
        y = self._get_y(i, j, x, xp)
        dy = xp[j] - y - 1
        fee = self.fee * dy // self.FEE_DENOMINATOR
        return (dy - fee) * self.PRECISION // rates[j]

    def _get_dy_underlying(
        self,
        i: int,
        j: int,
        dx: int,
        block_identifier: BlockIdentifier | None = None,
        override_state: CurveStableswapPoolState | None = None,
    ) -> int:
        if TYPE_CHECKING:
            assert self.base_pool is not None

        pool_balances = override_state.balances if override_state is not None else self.balances

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else get_number_for_block_identifier(
                identifier=block_identifier,
                w3=connection_manager.get_web3(self.chain_id),
            )
        )

        if self.address == "0x618788357D0EBd8A37e763ADab3bc575D54c2C7d":
            base_n_coins = len(self.base_pool.tokens)
            max_coin = len(self.tokens) - 1
            redemption_coin = 0

            # dx and dy in underlying units
            rates = (
                self._get_scaled_redemption_price(block_number=block_number),
                vp_rate := self._get_virtual_price(block_number=block_number),
            )
            xp = self._xp(rates=rates, balances=pool_balances)

            # Use base_i or base_j if they are >= 0
            base_i = i - max_coin
            base_j = j - max_coin
            meta_i = max_coin
            meta_j = max_coin
            if base_i < 0:
                meta_i = i
            if base_j < 0:
                meta_j = j

            if base_i < 0:
                x = xp[i] + (
                    dx
                    * self._get_scaled_redemption_price(block_number=block_number)
                    // self.PRECISION
                )
            elif base_j < 0:
                # i is from BasePool
                # At first, get the amount of pool tokens
                base_inputs = [0] * base_n_coins
                base_inputs[base_i] = dx
                # Token amount transformed to underlying "dollars"
                x = (
                    self.base_pool.calc_token_amount(
                        amounts=base_inputs,
                        deposit=True,
                        block_identifier=block_number,
                        override_state=(
                            override_state.base if override_state is not None else None
                        ),
                    )
                    * vp_rate
                    // self.PRECISION
                )
                # Accounting for deposit/withdraw fees approximately
                x -= x * self.base_pool.fee // (2 * self.FEE_DENOMINATOR)
                # Adding number of pool tokens
                x += xp[max_coin]
            else:
                # If both are from the base pool
                return self.base_pool.get_dy(
                    i=base_i,
                    j=base_j,
                    dx=dx,
                    override_state=(override_state.base if override_state is not None else None),
                )

            # This pool is involved only when in-pool assets are used
            y = self._get_y(meta_i, meta_j, x, xp)
            dy = xp[meta_j] - y - 1
            dy = dy - self.fee * dy // self.FEE_DENOMINATOR
            if j == redemption_coin:
                dy = (dy * self.PRECISION) // self._get_scaled_redemption_price(
                    block_number=block_number
                )

            # If output is going via the metapool
            if base_j >= 0:
                # j is from BasePool
                # The fee is already accounted for
                dy, *_ = self.base_pool.calc_withdraw_one_coin(
                    _token_amount=dy * self.PRECISION // vp_rate,
                    i=base_j,
                    block_identifier=block_number,
                )

            return dy

        if self.address in (
            "0xC61557C5d177bd7DC889A3b621eEC333e168f68A",
            "0x4606326b4Db89373F5377C316d3b0F6e55Bc6A20",
        ):
            base_n_coins = len(self.base_pool.tokens)
            max_coin = len(self.tokens) - 1

            rates = (self.PRECISION, self._get_virtual_price(block_number=block_number))
            xp = self._xp(rates=rates, balances=pool_balances)

            base_i = 0
            base_j = 0
            meta_i = 0
            meta_j = 0

            if i != 0:
                base_i = i - max_coin
                meta_i = 1
            if j != 0:
                base_j = j - max_coin
                meta_j = 1

            if i == 0:
                x = xp[i] + dx * (rates[0] // 10**18)
            elif j == 0:
                # i is from BasePool
                # At first, get the amount of pool tokens
                base_inputs = [0] * base_n_coins
                base_inputs[base_i] = dx
                # Token amount transformed to underlying "dollars"
                x = (
                    self.base_pool.calc_token_amount(
                        amounts=base_inputs,
                        deposit=True,
                        block_identifier=block_number,
                        override_state=(
                            override_state.base if override_state is not None else None
                        ),
                    )
                    * rates[1]
                    // self.PRECISION
                )
                # Accounting for deposit/withdraw fees approximately
                x -= x * self.base_pool.fee // (2 * self.FEE_DENOMINATOR)
                # Adding number of pool tokens
                x += xp[max_coin]
            else:
                # If both are from the base pool
                return self.base_pool.get_dy(
                    i=base_i,
                    j=base_j,
                    dx=dx,
                    block_identifier=block_number,
                    override_state=(override_state.base if override_state is not None else None),
                )

            # This pool is involved only when in-pool assets are used
            y = self._get_y(meta_i, meta_j, x, xp)
            dy = xp[meta_j] - y - 1
            dy = dy - self.fee * dy // self.FEE_DENOMINATOR

            # If output is going via the metapool
            if j == 0:
                dy //= rates[0] // 10**18
            else:
                # j is from BasePool
                # The fee is already accounted for
                dy, *_ = self.base_pool.calc_withdraw_one_coin(
                    _token_amount=dy * self.PRECISION // rates[1],
                    i=base_j,
                    block_identifier=block_number,
                )

            return dy

        _rates = list(self.rate_multipliers)

        vp_rate = self._get_virtual_price(block_number=block_number)
        _rates[-1] = vp_rate

        xp = self._xp(rates=tuple(_rates), balances=pool_balances)
        precisions = self.precision_multipliers

        base_n_coins = len(self.base_pool.tokens)
        max_coin = len(self.tokens) - 1

        # Use base_i or base_j if they are >= 0
        base_i = i - max_coin
        base_j = j - max_coin
        meta_i = max_coin
        meta_j = max_coin
        if base_i < 0:
            meta_i = i
        if base_j < 0:
            meta_j = j

        if base_i < 0:
            x = xp[i] + dx * precisions[i]
        elif base_j < 0:
            # i is from BasePool
            # At first, get the amount of pool tokens
            base_inputs = [0] * base_n_coins
            base_inputs[base_i] = dx
            # Token amount transformed to underlying "dollars"
            x = (
                self.base_pool.calc_token_amount(
                    amounts=base_inputs,
                    deposit=True,
                    block_identifier=block_number,
                    override_state=(override_state.base if override_state is not None else None),
                )
                * vp_rate
                // self.PRECISION
            )
            # Accounting for deposit/withdraw fees approximately
            x -= x * self.base_pool.fee // (2 * self.FEE_DENOMINATOR)
            # Adding number of pool tokens
            x += xp[max_coin]
        else:
            # If both are from the base pool
            return self.base_pool.get_dy(
                i=base_i,
                j=base_j,
                dx=dx,
                block_identifier=block_number,
                override_state=(override_state.base if override_state is not None else None),
            )

        # This pool is involved only when in-pool assets are used
        y = self._get_y(meta_i, meta_j, x, xp)
        dy = xp[meta_j] - y - 1
        dy = dy - self.fee * dy // self.FEE_DENOMINATOR

        # If output is going via the metapool
        if base_j < 0:
            dy //= precisions[meta_j]
        else:
            # j is from BasePool
            # The fee is already accounted for
            dy, *_ = self.base_pool.calc_withdraw_one_coin(
                _token_amount=dy * self.PRECISION // vp_rate,
                i=base_j,
                block_identifier=block_number,
            )

        return dy

    def _get_base_cache_updated(self, block_number: BlockNumber) -> int:
        with contextlib.suppress(KeyError):
            return self._cached_base_cache_updated[block_number]

        w3 = connection_manager.get_web3(self.chain_id)

        base_cache_updated: int
        (base_cache_updated,) = eth_abi.abi.decode(
            types=["uint256"],
            data=w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="base_cache_updated()")[:4],
                },
                block_identifier=block_number,
            ),
        )
        self._cached_base_cache_updated[block_number] = base_cache_updated
        return base_cache_updated

    def _get_base_virtual_price(self, block_number: BlockNumber) -> int:
        with contextlib.suppress(KeyError):
            return self._cached_base_virtual_price[block_number]

        w3 = connection_manager.get_web3(self.chain_id)

        base_virtual_price: int
        (base_virtual_price,) = eth_abi.abi.decode(
            types=["uint256"],
            data=w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="base_virtual_price()")[:4],
                },
                block_identifier=block_number,
            ),
        )
        self._cached_base_virtual_price[block_number] = base_virtual_price
        return base_virtual_price

    def _get_virtual_price(self, block_number: BlockNumber) -> int:
        if TYPE_CHECKING:
            assert self.base_pool is not None

        with contextlib.suppress(KeyError):
            return self._cached_virtual_price[block_number]

        base_cache_expires = 10 * 60  # 10 minutes

        w3 = connection_manager.get_web3(self.chain_id)
        self._block_timestamps[block_number] = w3.eth.get_block(block_identifier=block_number)[
            "timestamp"
        ]

        base_virtual_price: int
        if (
            self.base_cache_updated is None
            or self._block_timestamps[block_number] > self.base_cache_updated + base_cache_expires
        ):
            (base_virtual_price,) = eth_abi.abi.decode(
                types=["uint256"],
                data=w3.eth.call(
                    transaction={
                        "to": self.base_pool.address,
                        "data": Web3.keccak(text="get_virtual_price()")[:4],
                    },
                    block_identifier=block_number,
                ),
            )
        else:
            base_virtual_price = self.base_virtual_price

        self._cached_virtual_price[block_number] = base_virtual_price
        self.base_virtual_price = base_virtual_price
        return base_virtual_price

    def _get_admin_balances(self, block_number: BlockNumber) -> tuple[int, ...]:
        with contextlib.suppress(KeyError):
            return self._cached_admin_balances[block_number]

        admin_balances: list[int] = []
        for token_index, _ in enumerate(self.tokens):
            admin_balance: int
            (admin_balance,) = raw_call(
                w3=connection_manager.get_web3(chain_id=self.chain_id),
                address=self.address,
                calldata=encode_function_calldata(
                    function_prototype="admin_balances(uint256)",
                    function_arguments=[token_index],
                ),
                return_types=["uint256"],
                block_identifier=block_number,
            )
            admin_balances.append(admin_balance)

        self._cached_admin_balances[block_number] = tuple(admin_balances)
        return tuple(admin_balances)

    def _get_d(self, _xp: Sequence[int], _amp: int) -> int:
        """
        Solve for the Curve stableswap invariant D, using a modified Newton's method.

        Mainnet V1 Curve pools have several calculation variants to calculate the D and D_prev
        values. The pool addresses using each variant are grouped and the appropriate function is
        set at runtime.
        """

        def calc_d(
            *,
            a_nn: int,
            s: int,
            d: int,
            d_p: int,
            n_coins: int,
            a_precision: int,
        ) -> int:
            return (
                (a_nn * s // a_precision + d_p * n_coins)
                * d
                // ((a_nn - a_precision) * d // a_precision + (n_coins + 1) * d_p)
            )

        def calc_d_variant_alpha(
            *,
            a_nn: int,
            s: int,
            d: int,
            d_p: int,
            n_coins: int,
            a_precision: int,  # noqa:ARG001
        ) -> int:
            return (a_nn * s + d_p * n_coins) * d // ((a_nn - 1) * d + (n_coins + 1) * d_p)

        def calc_dp(
            *,
            d: int,
            d_p: int,
            xp: Sequence[int],
        ) -> int:
            for x in xp:
                d_p = d_p * d // (x * n_coins)
            return d_p

        def calc_dp_variant_alpha(
            *,
            d: int,
            d_p: int,
            xp: Sequence[int],
        ) -> int:
            for x in xp:
                d_p = d_p * d // (x * n_coins + 1)
            return d_p

        def calc_dp_variant_beta(
            *,
            d: int,
            d_p: int,  # noqa:ARG001
            xp: Sequence[int],
        ) -> int:
            return d * d // xp[0] * d // xp[1] // n_coins**2

        def calc_dp_variant_gamma(
            *,
            d: int,
            d_p: int,  # noqa:ARG001
            xp: Sequence[int],
        ) -> int:
            return d * d // xp[0] * d // xp[1] // cast("int", n_coins**n_coins)

        d_func = calc_d
        dp_func = calc_dp
        if self.address in self.D_VARIANT_GROUP_0:
            d_func = calc_d_variant_alpha
        elif self.address in self.D_VARIANT_GROUP_1:
            d_func = calc_d_variant_alpha
            dp_func = calc_dp_variant_alpha
        elif self.address in self.D_VARIANT_GROUP_2:
            dp_func = calc_dp_variant_beta
        elif self.address in self.D_VARIANT_GROUP_3:
            dp_func = calc_dp_variant_alpha
        elif self.address in self.D_VARIANT_GROUP_4:
            dp_func = calc_dp_variant_gamma

        d = s = sum(_xp)
        if s == 0:
            return 0
        n_coins = len(self.tokens)
        a_nn = _amp * n_coins

        for _ in range(255):  # pragma: no branch
            d_p = d_prev = d
            d_p = dp_func(d=d, d_p=d_p, xp=_xp)
            d = d_func(a_nn=a_nn, s=s, d=d, d_p=d_p, n_coins=n_coins, a_precision=self.A_PRECISION)
            if d_prev < d:
                if d - d_prev <= 1:
                    return d
            elif d_prev - d <= 1:
                return d

        raise EVMRevertError(error="D calculation did not converge.")  # pragma: no cover

    def _get_y(self, i: int, j: int, x: int, xp: Sequence[int]) -> int:
        """
        Calculate x[j] if one makes x[i] = x

        Done by solving quadratic equation iteratively.
        x_1**2 + x_1 * (sum' - (A*n**n - 1) * D / (A * n**n)) = D ** (n + 1) / (
            n ** (2 * n) * prod' * A
        )
        x_1**2 + b*x_1 = c

        x_1 = (x_1**2 + c) / (2*x_1 + b)
        """

        # x in the input is converted to the same price/precision

        n_coins = len(self.tokens)

        assert i != j, "same coin"
        assert j >= 0, "j below zero"
        assert j < n_coins, "j above N_COINS"

        # should be unreachable, but good for safety
        assert i >= 0
        assert i < n_coins

        amp = (
            self._a(timestamp=self._block_timestamps[self.update_block]) // self.A_PRECISION
            if self.address in self.Y_VARIANT_GROUP_0
            else self._a(timestamp=self._block_timestamps[self.update_block])
        )
        c = y = d = self._get_d(xp, amp)

        s = 0
        for coin_index in range(n_coins):
            if coin_index == i:
                _x = x
            elif coin_index != j:
                _x = xp[coin_index]
            else:
                continue
            s += _x
            c = c * d // (_x * n_coins)

        a_nn = amp * n_coins
        if self.address in self.Y_VARIANT_GROUP_1:
            c = c * d // (a_nn * n_coins)
            b = s + d // a_nn
        else:
            c = c * d * self.A_PRECISION // (a_nn * n_coins)
            b = s + d * self.A_PRECISION // a_nn

        for _ in range(255):  # pragma: no branch
            y_prev = y
            y = (y * y + c) // (2 * y + b - d)
            if y > y_prev:
                if y - y_prev <= 1:
                    return y
            elif y_prev - y <= 1:
                return y

        raise EVMRevertError(error="y calculation did not converge.")  # pragma: no cover

    def _get_y_d(self, a: int, i: int, xp: Sequence[int], d: int) -> int:
        n_coins = len(self.tokens)

        assert i >= 0  # dev: i below zero
        assert i < n_coins  # dev: i above N_COINS

        c = y = d

        s = 0
        for coin_index in range(n_coins):
            if coin_index != i:
                x = xp[coin_index]
            else:
                continue
            s += x
            c = c * d // (x * n_coins)

        a_nn = a * n_coins
        if self.address in self.Y_D_VARIANT_GROUP_0:
            b = s + d * self.A_PRECISION // a_nn
            c = c * d * self.A_PRECISION // (a_nn * n_coins)
        else:
            b = s + d // a_nn
            c = c * d // (a_nn * n_coins)

        for _ in range(255):  # pragma: no branch
            y_prev = y
            y = (y * y + c) // (2 * y + b - d)
            if y > y_prev:
                if y - y_prev <= 1:
                    return y
            elif y_prev - y <= 1:
                return y

        raise EVMRevertError(error="y_d calculation did not converge.")  # pragma: no cover

    def _set_oracle_method(self, block_number: BlockNumber) -> None:
        w3 = connection_manager.get_web3(self.chain_id)
        (self.oracle_method,) = eth_abi.abi.decode(
            types=["uint256"],
            data=w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="oracle_method()")[:4],
                },
                block_identifier=block_number,
            ),
        )

    def _stored_rates_from_ctokens(self, block_number: BlockNumber) -> tuple[int, ...]:
        with contextlib.suppress(KeyError):
            return self._cached_rates_from_ctokens[block_number]

        result: list[int] = []
        rate: int
        for token, use_lending, multiplier in zip(
            self.tokens,
            self.use_lending,
            self.precision_multipliers,
            strict=True,
        ):
            if not use_lending:
                rate = self.PRECISION
            else:
                w3 = connection_manager.get_web3(self.chain_id)
                (rate,) = eth_abi.abi.decode(
                    types=["uint256"],
                    data=w3.eth.call(
                        transaction={
                            "to": token.address,
                            "data": Web3.keccak(text="exchangeRateStored()")[:4],
                        },
                        block_identifier=block_number,
                    ),
                )
                supply_rate: int
                (supply_rate,) = eth_abi.abi.decode(
                    types=["uint256"],
                    data=w3.eth.call(
                        transaction={
                            "to": token.address,
                            "data": Web3.keccak(text="supplyRatePerBlock()")[:4],
                        },
                        block_identifier=block_number,
                    ),
                )
                old_block: BlockNumber
                (old_block,) = eth_abi.abi.decode(
                    types=["uint256"],
                    data=w3.eth.call(
                        transaction={
                            "to": token.address,
                            "data": Web3.keccak(text="accrualBlockNumber()")[:4],
                        },
                        block_identifier=block_number,
                    ),
                )

                rate += rate * supply_rate * (block_number - old_block) // self.PRECISION

            result.append(multiplier * rate)

        self._cached_rates_from_ctokens[block_number] = tuple(result)
        return tuple(result)

    def _stored_rates_from_ytokens(self, block_number: BlockNumber) -> tuple[int, ...]:
        with contextlib.suppress(KeyError):
            return self._cached_rates_from_ytokens[block_number]

        # ref: https://etherscan.io/address/0x79a8C46DeA5aDa233ABaFFD40F3A0A2B1e5A4F27#code

        result: list[int] = []
        for token, multiplier, use_lending in zip(
            self.tokens,
            self.precision_multipliers,
            self.use_lending,
            strict=True,
        ):
            if use_lending:
                w3 = connection_manager.get_web3(self.chain_id)
                rate: int
                (rate,) = eth_abi.abi.decode(
                    types=["uint256"],
                    data=w3.eth.call(
                        transaction={
                            "to": token.address,
                            "data": Web3.keccak(text="getPricePerFullShare()")[:4],
                        },
                        block_identifier=block_number,
                    ),
                )
            else:
                rate = self.LENDING_PRECISION

            result.append(rate * multiplier)

        self._cached_rates_from_ytokens[block_number] = tuple(result)
        return tuple(result)

    def _stored_rates_from_cytokens(self, block_number: BlockNumber) -> tuple[int, ...]:
        with contextlib.suppress(KeyError):
            return self._cached_rates_from_cytokens[block_number]

        w3 = connection_manager.get_web3(self.chain_id)

        result: list[int] = []
        for token, precision_multiplier in zip(
            self.tokens, self.precision_multipliers, strict=True
        ):
            rate: int
            (rate,) = eth_abi.abi.decode(
                types=["uint256"],
                data=(
                    w3.eth.call(
                        transaction={
                            "to": token.address,
                            "data": Web3.keccak(text="exchangeRateStored()")[:4],
                        },
                        block_identifier=block_number,
                    )
                ),
            )
            supply_rate: int
            (supply_rate,) = eth_abi.abi.decode(
                types=["uint256"],
                data=w3.eth.call(
                    transaction={
                        "to": token.address,
                        "data": Web3.keccak(text="supplyRatePerBlock()")[:4],
                    },
                    block_identifier=block_number,
                ),
            )
            old_block: BlockNumber
            (old_block,) = eth_abi.abi.decode(
                types=["uint256"],
                data=w3.eth.call(
                    transaction={
                        "to": token.address,
                        "data": Web3.keccak(text="accrualBlockNumber()")[:4],
                    },
                    block_identifier=block_number,
                ),
            )

            rate += rate * supply_rate * (block_number - old_block) // self.PRECISION
            result.append(precision_multiplier * rate)

        self._cached_rates_from_cytokens[block_number] = tuple(result)
        return tuple(result)

    def _stored_rates_from_reth(self, block_number: BlockNumber) -> tuple[int, ...]:
        with contextlib.suppress(KeyError):
            return self.PRECISION, self._cached_rates_from_reth[block_number]

        w3 = connection_manager.get_web3(self.chain_id)

        # ref: https://etherscan.io/address/0xF9440930043eb3997fc70e1339dBb11F341de7A8#code
        ratio: int
        (ratio,) = eth_abi.abi.decode(
            types=["uint256"],
            data=w3.eth.call(
                transaction={
                    "to": self.tokens[1].address,
                    "data": Web3.keccak(text="getExchangeRate()")[:4],
                },
                block_identifier=block_number,
            ),
        )
        self._cached_rates_from_reth[block_number] = ratio
        return self.PRECISION, ratio

    def _stored_rates_from_aeth(self, block_number: BlockNumber) -> tuple[int, ...]:
        with contextlib.suppress(KeyError):
            return (
                self.PRECISION,
                self.PRECISION
                * self.LENDING_PRECISION
                // self._cached_rates_from_aeth[block_number],
            )

        w3 = connection_manager.get_web3(self.chain_id)

        # ref: https://etherscan.io/address/0xA96A65c051bF88B4095Ee1f2451C2A9d43F53Ae2#code
        ratio: int
        (ratio,) = eth_abi.abi.decode(
            types=["uint256"],
            data=w3.eth.call(
                transaction={
                    "to": self.tokens[1].address,
                    "data": Web3.keccak(text="ratio()")[:4],
                },
                block_identifier=block_number,
            ),
        )
        self._cached_rates_from_aeth[block_number] = ratio
        return self.PRECISION, self.PRECISION * self.LENDING_PRECISION // ratio

    def _stored_rates_from_oracle(self, block_number: BlockNumber) -> tuple[int, ...]:
        """
        Get rates from on-chain oracle

        Ref: https://etherscan.io/address/0x59Ab5a5b5d617E478a2479B0cAD80DA7e2831492#code
        """

        with contextlib.suppress(KeyError):
            return self._cached_rates_from_oracle[block_number]

        self._set_oracle_method(block_number=block_number)
        if TYPE_CHECKING:
            assert self.oracle_method is not None

        if self.oracle_method == 0:
            rates = self.rate_multipliers
        else:
            oracle_bit_mask = (2**32 - 1) * 256**28
            oracle_rate: int
            (oracle_rate,) = eth_abi.abi.decode(
                types=["uint256"],
                data=connection_manager.get_web3(self.chain_id).eth.call(
                    transaction={
                        "to": get_checksum_address(HexBytes(self.oracle_method % 2**160)),
                        "data": HexBytes(self.oracle_method & oracle_bit_mask),
                    },
                    block_identifier=block_number,
                ),
            )
            rates = (
                self.rate_multipliers[0],
                self.rate_multipliers[1] * oracle_rate // self.PRECISION,
            )

        self._cached_rates_from_oracle[block_number] = rates
        return rates

    def _xp(self, rates: Iterable[int], balances: Iterable[int]) -> tuple[int, ...]:
        return tuple(
            rate * balance // self.PRECISION for rate, balance in zip(rates, balances, strict=True)
        )

    def auto_update(self, block_number: BlockNumber | None = None) -> bool:
        """
        Retrieve and set updated balances from the contract
        """

        with self._state_lock:
            w3 = connection_manager.get_web3(self.chain_id)

            state_block = w3.eth.get_block("latest" if block_number is None else block_number)
            block_number = state_block["number"]

            token_balances = []
            token_balance: int
            for token_id, _ in enumerate(self.tokens):
                (token_balance,) = eth_abi.abi.decode(
                    types=[self._coin_index_type],
                    data=w3.eth.call(
                        transaction={
                            "to": self.address,
                            "data": Web3.keccak(text=f"balances({self._coin_index_type})")[:4]
                            + eth_abi.abi.encode(types=[self._coin_index_type], args=[token_id]),
                        },
                        block_identifier=block_number,
                    ),
                )
                token_balances.append(token_balance)

            if self.base_pool is not None:
                self.base_pool.auto_update(block_number=block_number)
                if self.base_cache_updated is not None:
                    self.base_cache_updated = self._get_base_cache_updated(
                        block_number=block_number
                    )

            found_updates = tuple(token_balances) != self.balances

            state = (
                CurveStableswapPoolState(
                    address=self.address,
                    balances=tuple(token_balances),
                    base=self.base_pool.state,
                    block=block_number,
                )
                if self.base_pool is not None
                else CurveStableswapPoolState(
                    address=self.address,
                    balances=tuple(token_balances),
                    block=block_number,
                )
            )
            self._state_cache[block_number] = state
            self._state = state
            self._block_timestamps[block_number] = state_block["timestamp"]

            if found_updates:
                self._notify_subscribers(
                    message=CurveStableSwapPoolStateUpdated(state),
                )

            return found_updates

    def calculate_tokens_out_from_tokens_in(
        self,
        token_in: Erc20Token,
        token_out: Erc20Token,
        token_in_quantity: int,
        override_state: CurveStableswapPoolState | None = None,
        block_identifier: BlockIdentifier | None = None,
    ) -> int:
        """
        Calculates the expected token OUTPUT for a target INPUT at current pool reserves.
        """

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else get_number_for_block_identifier(
                block_identifier,
                connection_manager.get_web3(self.chain_id),
            )
        )

        if token_in_quantity <= 0:
            raise InvalidSwapInputAmount

        if override_state:
            logger.debug("Overrides applied:")
            logger.debug(f"Balances: {override_state.balances}")

        tokens_used_this_pool = [
            token_in in self.tokens,
            token_out in self.tokens,
        ]

        tokens_used_in_base_pool = []
        if self.base_pool is not None:
            tokens_used_in_base_pool = [
                token_in in self.base_pool.tokens,
                token_out in self.base_pool.tokens,
            ]

        if all(tokens_used_this_pool):
            if any(balance == 0 for balance in self.balances):
                raise NoLiquidity(message="One or more of the tokens has a zero balance.")

            return self.get_dy(
                i=self.tokens.index(token_in),
                j=self.tokens.index(token_out),
                dx=token_in_quantity,
                block_identifier=block_number,
                override_state=override_state,
            )
        if any(tokens_used_this_pool) and any(tokens_used_in_base_pool):
            if TYPE_CHECKING:
                assert self.base_pool is not None

            # TODO: see if any of these checks are unnecessary (partial zero balance OK?)
            if any(balance == 0 for balance in self.base_pool.balances):
                raise NoLiquidity(message="One or more of the base pool tokens has a zero balance.")
            if any(balance == 0 for balance in self.balances):
                raise NoLiquidity(message="One or more of the tokens has a zero balance.")

            token_in_from_metapool = token_in in self.tokens
            token_out_from_metapool = token_out in self.tokens
            assert token_in_from_metapool or token_out_from_metapool

            if token_in_from_metapool and self.balances[self.tokens.index(token_in)] == 0:
                raise NoLiquidity(message=f"{token_in} has a zero balance.")
            if token_out_from_metapool and self.balances[self.tokens.index(token_out)] == 0:
                raise NoLiquidity(message=f"{token_out} has a zero balance.")

            token_in_from_basepool = token_in in self.base_pool.tokens
            token_out_from_basepool = token_out in self.base_pool.tokens
            assert token_in_from_basepool or token_out_from_basepool

            if (
                token_in_from_basepool
                and self.base_pool.balances[self.base_pool.tokens.index(token_in)] == 0
            ):
                raise NoLiquidity(message=f"{token_in} has a zero balance.")
            if (
                token_out_from_basepool
                and self.base_pool.balances[self.base_pool.tokens.index(token_out)] == 0
            ):
                raise NoLiquidity(message=f"{token_out} has a zero balance.")

            return self._get_dy_underlying(
                i=(
                    self.tokens.index(token_in)
                    if token_in_from_metapool
                    else self.tokens_underlying.index(token_in)
                ),
                j=(
                    self.tokens.index(token_out)
                    if token_out_from_metapool
                    else self.tokens_underlying.index(token_out)
                ),
                dx=token_in_quantity,
                block_identifier=block_number,
                override_state=override_state,
            )
        if all(tokens_used_in_base_pool):
            token_in_from_basepool = token_in in self.tokens_underlying
            token_out_from_basepool = token_out in self.tokens_underlying
            assert token_in_from_basepool or token_out_from_basepool

            return self._get_dy_underlying(
                i=self.tokens_underlying.index(token_in),
                j=self.tokens_underlying.index(token_out),
                dx=token_in_quantity,
                block_identifier=block_number,
                override_state=override_state,
            )

        raise DegenbotValueError(
            message="Tokens not held by pool or in underlying base pool"
        )  # pragma: no cover

    def get_arbitrage_helpers(self) -> tuple[AbstractArbitrage, ...]:
        return tuple(
            subscriber
            for subscriber in self._subscribers
            if isinstance(subscriber, AbstractArbitrage)
        )



from degenbot.exceptions.evm import EVMRevertError

ONE_18 = 1 * 10**18

# Internally, intermediate values are computed with higher precision as 20 decimal fixed point
# numbers, and in the case of ln36, 36 decimals.
ONE_20 = 1 * 10**20
ONE_36 = 1 * 10**36


# The domain of natural exponentiation is bound by the word size and number of decimals used.
#
# Because internally the result will be stored using 20 decimals, the largest possible result is
# (2^255 - 1) / 10^20, which makes the largest exponent
# ln((2^255 - 1) / 10^20) = 130.700829182905140221.
# The smallest possible result is 10^(-18), which makes largest negative argument
# ln(10^(-18)) = -41.446531673892822312.
# We use 130.0 and -41.0 to have some safety margin.
MAX_NATURAL_EXPONENT = 130 * 10**18
MIN_NATURAL_EXPONENT = -41 * 10**18


# Bounds for ln_36's argument. Both ln(0.9) and ln(1.1) can be represented with 36 decimal places in
# a fixed point 256 bit integer.
LN_36_LOWER_BOUND = ONE_18 - 1 * 10**17
LN_36_UPPER_BOUND = ONE_18 + 1 * 10**17
MILD_EXPONENT_BOUND = 2**254 // ONE_20

# 18 decimal constants
x0 = 128000000000000000000  # 2^7
a0 = 38877084059945950922200000000000000000000000000000000000  # e^(x0) (no decimals)
x1 = 64000000000000000000  # 2^6
a1 = 6235149080811616882910000000  # e^(x1) (no decimals)

# 20 decimal constants
x2 = 3200000000000000000000  # 2^5
a2 = 7896296018268069516100000000000000  # e^(x2)
x3 = 1600000000000000000000  # 2^4
a3 = 888611052050787263676000000  # e^(x3)
x4 = 800000000000000000000  # 2^3
a4 = 298095798704172827474000  # e^(x4)
x5 = 400000000000000000000  # 2^2
a5 = 5459815003314423907810  # e^(x5)
x6 = 200000000000000000000  # 2^1
a6 = 738905609893065022723  # e^(x6)
x7 = 100000000000000000000  # 2^0
a7 = 271828182845904523536  # e^(x7)
x8 = 50000000000000000000  # 2^-1
a8 = 164872127070012814685  # e^(x8)
x9 = 25000000000000000000  # 2^-2
a9 = 128402541668774148407  # e^(x9)
x10 = 12500000000000000000  # 2^-3
a10 = 113314845306682631683  # e^(x10)
x11 = 6250000000000000000  # 2^-4
a11 = 106449445891785942956  # e^(x11)


def pow(x: int, y: int) -> int:  # noqa: A001
    if y == 0:
        # We solve the 0^0 indetermination by making it equal one.
        return ONE_18

    if x == 0:
        return 0

    # Instead of computing x^y directly, we instead rely on the properties of logarithms and
    # exponentiation to arrive at that result. In particular, exp(ln(x)) = x, and
    # ln(x^y) = y * ln(x). This means x^y = exp(y * ln(x)).

    # The ln function takes a signed value, so we need to make sure x fits in the signed 256 bit
    # range.
    if x >= 2**255:
        raise EVMRevertError(error="X_OUT_OF_BOUNDS")
    x_int256 = x

    # We will compute y * ln(x) in a single step. Depending on the value of x, we can either use ln
    # or ln_36. In both cases, we leave the division by ONE_18 (due to fixed point multiplication)
    # to the end.

    # This prevents y * ln(x) from overflowing, and at the same time guarantees y fits in the signed
    # 256 bit range.
    if y >= MILD_EXPONENT_BOUND:
        raise EVMRevertError(error="Y_OUT_OF_BOUNDS")
    y_int256 = y

    if LN_36_LOWER_BOUND < x_int256 < LN_36_UPPER_BOUND:
        ln_36_x = _ln_36(x_int256)

        # ln_36_x has 36 decimal places, so multiplying by y_int256 isn't as straightforward, since
        # we can't just bring y_int256 to 36 decimal places, as it might overflow. Instead, we
        # perform two 18 decimal multiplications and add the results: one with the first 18 decimals
        # of ln_36_x, and one with the (downscaled) last 18 decimals.
        logx_times_y = (ln_36_x // ONE_18) * y_int256 + ((ln_36_x % ONE_18) * y_int256) // ONE_18
    else:
        logx_times_y = _ln(x_int256) * y_int256

    logx_times_y //= ONE_18

    # Finally, we compute exp(y * ln(x)) to arrive at x^y
    if not (MIN_NATURAL_EXPONENT <= logx_times_y <= MAX_NATURAL_EXPONENT):
        raise EVMRevertError(error="PRODUCT_OUT_OF_BOUNDS")

    return exp(logx_times_y)


def exp(x: int) -> int:
    if not (MIN_NATURAL_EXPONENT <= x <= MAX_NATURAL_EXPONENT):
        raise EVMRevertError(error="Invalid exponent")

    if x < 0:
        # We only handle positive exponents: e^(-x) is computed as 1 / e^x. We can safely make x
        # positive since it fits in the signed 256 bit range (as it is larger than
        # MIN_NATURAL_EXPONENT). Fixed point division requires multiplying by ONE_18.
        return (ONE_18 * ONE_18) // exp(-x)

    # First, we use the fact that e^(x+y) = e^x * e^y to decompose x into a sum of powers of two,
    # which we call x_n, where x_n == 2^(7 - n), and e^x_n = a_n has been precomputed. We choose the
    # first x_n, x0, to equal 2^7 because all larger powers are larger than MAX_NATURAL_EXPONENT,
    # and therefore not present in the decomposition.
    # At the end of this process we will have the product of all e^x_n = a_n that apply, and the
    # remainder of this decomposition, which will be lower than the smallest x_n.
    # exp(x) = k_0 * a_0 * k_1 * a_1 * ... + k_n * a_n * exp(remainder), where each k_n equals
    # either 0 or 1.
    # We mutate x by subtracting x_n, making it the remainder of the decomposition.

    # The first two a_n (e^(2^7) and e^(2^6)) are too large if stored as 18 decimal numbers, and
    # could cause intermediate overflows. Instead we store them as plain integers, with 0 decimals.
    # Additionally, x0 + x1 is larger than MAX_NATURAL_EXPONENT, which means they will not both be
    # present in the decomposition.

    # For each x_n, we test if that term is present in the decomposition (if x is larger than it),
    # and if so deduct it and compute the accumulated product.

    first_an: int
    if x >= x0:
        x -= x0
        first_an = a0
    elif x >= x1:
        x -= x1
        first_an = a1
    else:
        first_an = 1  # One with no decimal places

    # We now transform x into a 20 decimal fixed point number, to have enhanced precision when
    # computing the smaller terms.
    x *= 100

    # `product` is the accumulated product of all a_n (except a0 and a1), which starts at 20 decimal
    # fixed point one. Recall that fixed point multiplication requires dividing by ONE_20.
    product = ONE_20

    if x >= x2:
        x -= x2
        product = (product * a2) // ONE_20
    if x >= x3:
        x -= x3
        product = (product * a3) // ONE_20
    if x >= x4:
        x -= x4
        product = (product * a4) // ONE_20
    if x >= x5:
        x -= x5
        product = (product * a5) // ONE_20
    if x >= x6:
        x -= x6
        product = (product * a6) // ONE_20
    if x >= x7:
        x -= x7
        product = (product * a7) // ONE_20
    if x >= x8:
        x -= x8
        product = (product * a8) // ONE_20
    if x >= x9:
        x -= x9
        product = (product * a9) // ONE_20

    # x10 and x11 are unnecessary here since we have high enough precision already.

    # Now we need to compute e^x, where x is small (in particular, it is smaller than x9). We use
    # the Taylor series expansion for e^x: 1 + x + (x^2 / 2!) + (x^3 / 3!) + ... + (x^n / n!).

    series_sum = ONE_20  # The initial one in the sum, with 20 decimal places.
    term: int  # Each term in the sum, where the nth term is (x^n / n!).

    # The first term is simply x.
    term = x
    series_sum += term

    # Each term (x^n / n!) equals the previous one times x, divided by n. Since x is a fixed point
    # number, multiplying by it requires dividing by ONE_20, but dividing by the non-fixed point n
    # values does not.

    term = ((term * x) // ONE_20) // 2
    series_sum += term

    term = ((term * x) // ONE_20) // 3
    series_sum += term

    term = ((term * x) // ONE_20) // 4
    series_sum += term

    term = ((term * x) // ONE_20) // 5
    series_sum += term

    term = ((term * x) // ONE_20) // 6
    series_sum += term

    term = ((term * x) // ONE_20) // 7
    series_sum += term

    term = ((term * x) // ONE_20) // 8
    series_sum += term

    term = ((term * x) // ONE_20) // 9
    series_sum += term

    term = ((term * x) // ONE_20) // 10
    series_sum += term

    term = ((term * x) // ONE_20) // 11
    series_sum += term

    term = ((term * x) // ONE_20) // 12
    series_sum += term

    # 12 Taylor terms are sufficient for 18 decimal precision.

    # We now have the first a_n (with no decimals), and the product of all other a_n present, and
    # the Taylor approximation of the exponentiation of the remainder (both with 20 decimals). All
    # that remains is to multiply all three (one 20 decimal fixed point multiplication, dividing by
    # ONE_20, and one integer multiplication), and then drop two digits to return an 18 decimal
    # value.

    return (((product * series_sum) // ONE_20) * first_an) // 100


# @dev Logarithm (log(arg, base), with signed 18 decimal fixed point base and argument.
def log(arg: int, base: int) -> int:
    # This performs a simple base change: log(arg, base) = ln(arg) / ln(base).

    # Both logBase and logArg are computed as 36 decimal fixed point numbers, either by using ln_36,
    # or by upscaling.

    log_base = _ln_36(base) if LN_36_LOWER_BOUND < base < LN_36_UPPER_BOUND else _ln(base) * ONE_18
    log_arg = _ln_36(arg) if LN_36_LOWER_BOUND < arg < LN_36_UPPER_BOUND else _ln(arg) * ONE_18

    # When dividing, we multiply by ONE_18 to arrive at a result with 18 decimal places
    return (log_arg * ONE_18) // log_base


def ln(a: int) -> int:
    # The real natural logarithm is not defined for negative numbers or zero.
    if a <= 0:
        raise EVMRevertError(error="OUT_OF_BOUNDS")

    if LN_36_LOWER_BOUND < a < LN_36_UPPER_BOUND:
        return _ln_36(a) // ONE_18

    return _ln(a)


def _ln(a: int) -> int:
    if a < ONE_18:
        # Since ln(a^k) = k * ln(a), we can compute ln(a) as ln(a) = ln((1/a)^(-1)) = - ln((1/a)).
        # If a is less than one, 1/a will be greater than one, and this if statement will not be
        # entered in the recursive call. Fixed point division requires multiplying by ONE_18.
        return -_ln((ONE_18 * ONE_18) // a)

    # First, we use the fact that ln^(a * b) = ln(a) + ln(b) to decompose ln(a) into a sum of powers
    # of two, which we call x_n, where x_n == 2^(7 - n), which are the natural logarithm of
    # precomputed quantities a_n (that is, ln(a_n) = x_n). We choose the first x_n, x0, to equal 2^7
    # because the exponential of all larger powers cannot be represented as 18 fixed point decimal
    # numbers in 256 bits, and are therefore larger than a. At the end of this process we will have
    # the sum of all x_n = ln(a_n) that apply, and the remainder of this decomposition, which will
    # be lower than the smallest a_n.
    # ln(a) = k_0 * x_0 + k_1 * x_1 + ... + k_n * x_n + ln(remainder), where each k_n equals either
    # 0 or 1. We mutate a by subtracting a_n, making it the remainder of the decomposition.

    # For reasons related to how `exp` works, the first two a_n (e^(2^7) and e^(2^6)) are not stored
    # as fixed point numbers with 18 decimals, but instead as plain integers with 0 decimals, so we
    # need to multiply them by ONE_18 to convert them to fixed point. For each a_n, we test if that
    # term is present in the decomposition (if a is larger than it), and if so divide by it and
    # compute the accumulated sum.

    _sum = 0
    if a >= a0 * ONE_18:
        a //= a0  # Integer, not fixed point division
        _sum += x0

    if a >= a1 * ONE_18:
        a //= a1  # Integer, not fixed point division
        _sum += x1

    # All other a_n and x_n are stored as 20 digit fixed point numbers, so we convert the sum and a
    # to this format.
    _sum *= 100
    a *= 100

    # Because further a_n are  20 digit fixed point numbers, we multiply by ONE_20 when dividing by
    # them.

    if a >= a2:
        a = (a * ONE_20) // a2
        _sum += x2
    if a >= a3:
        a = (a * ONE_20) // a3
        _sum += x3
    if a >= a4:
        a = (a * ONE_20) // a4
        _sum += x4
    if a >= a5:
        a = (a * ONE_20) // a5
        _sum += x5
    if a >= a6:
        a = (a * ONE_20) // a6
        _sum += x6
    if a >= a7:
        a = (a * ONE_20) // a7
        _sum += x7
    if a >= a8:
        a = (a * ONE_20) // a8
        _sum += x8
    if a >= a9:
        a = (a * ONE_20) // a9
        _sum += x9
    if a >= a10:
        a = (a * ONE_20) // a10
        _sum += x10
    if a >= a11:
        a = (a * ONE_20) // a11
        _sum += x11

    # a is now a small number (smaller than a_11, which roughly equals 1.06). This means we can use
    # a Taylor series that converges rapidly for values of `a` close to one - the same one used in
    # ln_36.
    # Let z = (a - 1) / (a + 1).
    # ln(a) = 2 * (z + z^3 / 3 + z^5 / 5 + z^7 / 7 + ... + z^(2 * n + 1) / (2 * n + 1))

    # Recall that 20 digit fixed point division requires multiplying by ONE_20, and multiplication
    # requires division by ONE_20.
    z = ((a - ONE_20) * ONE_20) // (a + ONE_20)
    z_squared = (z * z) // ONE_20

    # num is the numerator of the series: the z^(2 * n + 1) term
    num = z

    # seriesSum holds the accumulated sum of each term in the series, starting with the initial z
    series_sum = num

    # In each step, the numerator is multiplied by z^2
    num = (num * z_squared) // ONE_20
    series_sum += num // 3

    num = (num * z_squared) // ONE_20
    series_sum += num // 5

    num = (num * z_squared) // ONE_20
    series_sum += num // 7

    num = (num * z_squared) // ONE_20
    series_sum += num // 9

    num = (num * z_squared) // ONE_20
    series_sum += num // 11

    # 6 Taylor terms are sufficient for 36 decimal precision.

    # Finally, we multiply by 2 (non fixed point) to compute ln(remainder)
    series_sum *= 2

    # We now have the sum of all x_n present, and the Taylor approximation of the logarithm of the
    # remainder (both with 20 decimals). All that remains is to sum these two, and then drop two
    # digits to return a 18 decimal value.

    return (_sum + series_sum) // 100


def _ln_36(x: int) -> int:
    # Since ln(1) = 0, a value of x close to one will yield a very small result, which makes using
    # 36 digits worthwhile.

    # First, we transform x to a 36 digit fixed point value.
    x *= ONE_18

    # We will use the following Taylor expansion, which converges very rapidly.
    # Let z = (x - 1) / (x + 1).
    # ln(x) = 2 * (z + z^3 / 3 + z^5 / 5 + z^7 / 7 + ... + z^(2 * n + 1) / (2 * n + 1))

    # Recall that 36 digit fixed point division requires multiplying by ONE_36, and multiplication
    # requires division by ONE_36.
    z = ((x - ONE_36) * ONE_36) // (x + ONE_36)
    z_squared = (z * z) // ONE_36

    # num is the numerator of the series: the z^(2 * n + 1) term
    num = z

    # seriesSum holds the accumulated sum of each term in the series, starting with the initial z
    series_sum = num

    # In each step, the numerator is multiplied by z^2
    num = (num * z_squared) // ONE_36
    series_sum += num // 3

    num = (num * z_squared) // ONE_36
    series_sum += num // 5

    num = (num * z_squared) // ONE_36
    series_sum += num // 7

    num = (num * z_squared) // ONE_36
    series_sum += num // 9

    num = (num * z_squared) // ONE_36
    series_sum += num // 11

    num = (num * z_squared) // ONE_36
    series_sum += num // 13

    num = (num * z_squared) // ONE_36
    series_sum += num // 15

    # 8 Taylor terms are sufficient for 36 decimal precision.

    # All that remains is multiplying by 2 (non fixed point).
    return series_sum * 2



import functools
from typing import overload

from pydantic import validate_call

from degenbot.constants import MAX_UINT160, MAX_UINT256, MIN_UINT160
from degenbot.exceptions.evm import EVMRevertError
from degenbot.uniswap.v4_libraries._config import V4_LIB_CACHE_SIZE
from degenbot.uniswap.v4_libraries.fixed_point_96 import Q96, RESOLUTION
from degenbot.uniswap.v4_libraries.full_math import muldiv, muldiv_rounding_up
from degenbot.uniswap.v4_libraries.functions import mulmod
from degenbot.uniswap.v4_libraries.unsafe_math import div_rounding_up
from degenbot.validation.evm_values import (
    ValidatedInt128,
    ValidatedInt256,
    ValidatedUint128,
    ValidatedUint128NonZero,
    ValidatedUint160,
    ValidatedUint160NonZero,
    ValidatedUint256,
)


@overload
def get_amount0_delta(
    *,
    sqrt_price_a_x96: ValidatedUint160,
    sqrt_price_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128,
    round_up: bool,
) -> ValidatedUint256: ...


@overload
def get_amount0_delta(
    *,
    sqrt_price_a_x96: ValidatedUint160,
    sqrt_price_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128,
    round_up: None,
) -> ValidatedInt256: ...


@functools.lru_cache(maxsize=V4_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_amount0_delta(
    *,
    sqrt_price_a_x96: ValidatedUint160NonZero,
    sqrt_price_b_x96: ValidatedUint160,
    liquidity: "ValidatedInt128 | ValidatedUint128",
    round_up: bool | None = None,
) -> "ValidatedInt256 | ValidatedUint256":
    """
    Gets the amount0 delta between two prices
    """

    if round_up is None:
        return (
            get_amount0_delta(
                sqrt_price_a_x96=sqrt_price_a_x96,
                sqrt_price_b_x96=sqrt_price_b_x96,
                liquidity=-liquidity,
                round_up=False,
            )
            if liquidity < 0
            else -get_amount0_delta(
                sqrt_price_a_x96=sqrt_price_a_x96,
                sqrt_price_b_x96=sqrt_price_b_x96,
                liquidity=liquidity,
                round_up=True,
            )
        )

    if sqrt_price_a_x96 > sqrt_price_b_x96:
        sqrt_price_a_x96, sqrt_price_b_x96 = sqrt_price_b_x96, sqrt_price_a_x96

    numerator1 = liquidity << RESOLUTION
    numerator2 = sqrt_price_b_x96 - sqrt_price_a_x96
    return (
        div_rounding_up(
            muldiv_rounding_up(numerator1, numerator2, sqrt_price_b_x96), sqrt_price_a_x96
        )
        if round_up
        else muldiv(numerator1, numerator2, sqrt_price_b_x96) // sqrt_price_a_x96
    )


@overload
def get_amount1_delta(
    *,
    sqrt_price_a_x96: ValidatedUint160,
    sqrt_price_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128,
    round_up: bool,
) -> ValidatedUint256: ...


@overload
def get_amount1_delta(
    *,
    sqrt_price_a_x96: ValidatedUint160,
    sqrt_price_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128,
    round_up: None,
) -> ValidatedInt256: ...


@functools.lru_cache(maxsize=V4_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_amount1_delta(
    *,
    sqrt_price_a_x96: ValidatedUint160,
    sqrt_price_b_x96: ValidatedUint160,
    liquidity: "ValidatedInt128 | ValidatedUint128",
    round_up: bool | None = None,
) -> "ValidatedInt256 | ValidatedUint256":
    """
    Gets the amount1 delta between two prices
    """

    if round_up is None:
        return (
            get_amount1_delta(
                sqrt_price_a_x96=sqrt_price_a_x96,
                sqrt_price_b_x96=sqrt_price_b_x96,
                liquidity=-liquidity,
                round_up=False,
            )
            if liquidity < 0
            else -get_amount1_delta(
                sqrt_price_a_x96=sqrt_price_a_x96,
                sqrt_price_b_x96=sqrt_price_b_x96,
                liquidity=liquidity,
                round_up=True,
            )
        )

    numerator = abs(sqrt_price_a_x96 - sqrt_price_b_x96)
    denominator = Q96
    _liquidity = liquidity
    # Equivalent to:
    # ... amount1 = roundUp
    #       ? FullMath.mulDivRoundingUp(liquidity, sqrtPriceBX96 - sqrtPriceAX96, FixedPoint96.Q96)
    #       : FullMath.mulDiv(liquidity, sqrtPriceBX96 - sqrtPriceAX96, FixedPoint96.Q96);
    # Cannot overflow because `type(uint128).max * type(uint160).max >> 96 < (1 << 192)`.
    return muldiv(_liquidity, numerator, denominator) + (
        int(mulmod(_liquidity, numerator, denominator) > 0) & round_up
    )


@functools.lru_cache(maxsize=V4_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_next_sqrt_price_from_amount0_rounding_up(
    *,
    sqrt_price_x96: ValidatedUint160,
    liquidity: ValidatedUint128,
    amount: ValidatedUint256,
    add: bool,
) -> ValidatedUint160:
    """
    Gets the next sqrt price given a delta of currency0
    """

    # Short circuit amount == 0 because the result is otherwise not guaranteed to equal the input
    # price
    if amount == 0:
        return sqrt_price_x96

    numerator1 = liquidity << RESOLUTION
    product = amount * sqrt_price_x96

    # @dev the Solidity contract uses an unchecked math block to determine if overflow has occured.
    # Python integer math cannot overflow, so check directly if the numerator exceeds the uint256
    # upper bound
    if add:
        if product < MAX_UINT256:  # product did not overflow
            denominator = numerator1 + product
            if denominator >= numerator1:
                # always fits in 160 bits
                return muldiv_rounding_up(
                    a=numerator1,
                    b=sqrt_price_x96,
                    denominator=denominator,
                )

        # product overflowed - failsafe path
        result = div_rounding_up(
            x=numerator1,
            y=(numerator1 // sqrt_price_x96) + amount,
        )
        assert MIN_UINT160 <= result <= MAX_UINT160
        return result

    # @dev: removed overflow check on denominator in favor of input validation on muldiv_rounding_up
    denominator = numerator1 - product
    return muldiv_rounding_up(numerator1, sqrt_price_x96, denominator)


@functools.lru_cache(maxsize=V4_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_next_sqrt_price_from_amount1_rounding_down(
    *,
    sqrt_price_x96: ValidatedUint160,
    liquidity: ValidatedUint128,
    amount: ValidatedUint256,
    add: bool,
) -> ValidatedUint160:
    """
    Gets the next sqrt price given a delta of currency1
    """

    # if we're adding (subtracting), rounding down requires rounding the quotient down (up)
    # in both cases, avoid a mulDiv for most inputs
    if add:
        quotient = (
            (amount << RESOLUTION) // liquidity
            if amount <= MAX_UINT160
            else muldiv(amount, Q96, liquidity)
        )
        return sqrt_price_x96 + quotient

    quotient = (
        div_rounding_up(amount << RESOLUTION, liquidity)
        if amount <= MAX_UINT160
        else muldiv_rounding_up(amount, Q96, liquidity)
    )

    if sqrt_price_x96 <= quotient:
        raise EVMRevertError(error="NotEnoughLiquidity")

    # always fits 160 bits
    return sqrt_price_x96 - quotient


@functools.lru_cache(maxsize=V4_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_next_sqrt_price_from_input(
    *,
    sqrt_price_x96: ValidatedUint160NonZero,
    liquidity: ValidatedUint128NonZero,
    amount_in: ValidatedUint256,
    zero_for_one: bool,
) -> ValidatedUint160:
    """
    Gets the next sqrt price given an input amount of currency0 or currency1, rounding to ensure
    that the target price is not passed.
    """

    return (
        get_next_sqrt_price_from_amount0_rounding_up(
            sqrt_price_x96=sqrt_price_x96,
            liquidity=liquidity,
            amount=amount_in,
            add=True,
        )
        if zero_for_one
        else get_next_sqrt_price_from_amount1_rounding_down(
            sqrt_price_x96=sqrt_price_x96,
            liquidity=liquidity,
            amount=amount_in,
            add=True,
        )
    )


@functools.lru_cache(maxsize=V4_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_next_sqrt_price_from_output(
    *,
    sqrt_price_x96: ValidatedUint160NonZero,
    liquidity: ValidatedUint128NonZero,
    amount_out: ValidatedUint256,
    zero_for_one: bool,
) -> ValidatedUint160:
    """
    Gets the next sqrt price given an output amount of currency0 or currency1, rounding to ensure
    that the target price is not passed.
    """

    return (
        get_next_sqrt_price_from_amount1_rounding_down(
            sqrt_price_x96=sqrt_price_x96,
            liquidity=liquidity,
            amount=amount_out,
            add=False,
        )
        if zero_for_one
        else get_next_sqrt_price_from_amount0_rounding_up(
            sqrt_price_x96=sqrt_price_x96,
            liquidity=liquidity,
            amount=amount_out,
            add=False,
        )
    )



import json
from collections.abc import Iterable
from typing import TYPE_CHECKING, Annotated, Any, Literal

import aiohttp
import eth_account
import pydantic
from eth_typing import HexStr
from hexbytes import HexBytes

from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.base import ExternalServiceError
from degenbot.functions import eip_191_hash
from degenbot.types.aliases import BlockNumber

if TYPE_CHECKING:
    from eth_account.signers.local import LocalAccount


class BuilderEndpoint:  # pragma: no cover
    """
    An external HTTP endpoint with one or more bundle-related methods defined by the Flashbots RPC
    specification at https://docs.flashbots.net/flashbots-auction/advanced/rpc-endpoint
    """

    def __init__(
        self,
        url: str,
        endpoints: Iterable[str],
        authentication_header_label: str | None = None,
    ) -> None:
        if not url.startswith(("http://", "https://")):
            raise DegenbotValueError(message="Invalid URL")
        self.url = url
        self.endpoints = tuple(endpoints)
        self.authentication_header_label = authentication_header_label

    @staticmethod
    def _build_authentication_header(
        payload: str,
        header_label: str | None,
        signer_key: str | None,
    ) -> dict[str, Any]:
        """
        Build a MIME type header with an optional EIP-191 signature of the provided payload.
        """

        bundle_headers = {"Content-Type": "application/json"}

        if signer_key and header_label:
            signer_account: LocalAccount = eth_account.Account.from_key(signer_key)
            signer_address = signer_account.address
            message_signature = eip_191_hash(message=payload, private_key=signer_key)
            bundle_signature = f"{signer_address}:{message_signature}"
            bundle_headers[header_label] = bundle_signature

        return bundle_headers

    @staticmethod
    async def send_payload(
        url: str,
        headers: dict[str, Any],
        data: str,
        http_session: aiohttp.ClientSession | None = None,
    ) -> Any:
        close_session_after_post = http_session is None

        session = (
            aiohttp.ClientSession(raise_for_status=True) if http_session is None else http_session
        )

        try:
            async with session.post(
                url=url,
                data=data,
                headers=headers,
            ) as resp:
                relay_response = await resp.json(
                    # Some builders omit or return an invalid MIME type, so use None to bypass the
                    # check in the `json` method
                    content_type=None,
                )
        except aiohttp.ClientError as exc:
            raise ExternalServiceError(error=str(exc)) from exc
        else:
            return (
                relay_response["error"] if "error" in relay_response else relay_response["result"]
            )
        finally:
            if close_session_after_post:
                await session.close()

    async def call_eth_bundle(
        self,
        bundle: Iterable[HexBytes],
        block_number: BlockNumber,
        state_block: BlockNumber | Literal["latest"],
        signer_key: str,
        block_timestamp: int | None = None,
        http_session: aiohttp.ClientSession | None = None,
    ) -> Any:
        """
        Send a formatted bundle to the eth_callBundle endpoint for simulation against state at a
        given block number
        """

        class EthCallBundleParams(pydantic.BaseModel):
            txs: list[HexStr]
            block: Annotated[
                int | Literal["latest"],
                pydantic.PlainSerializer(
                    lambda block_number: hex(block_number)
                    if isinstance(block_number, int)
                    else block_number,
                    return_type=HexStr,
                    when_used="json",
                ),
            ] = pydantic.Field(serialization_alias="blockNumber")
            state_block: Annotated[
                int | Literal["latest"],
                pydantic.PlainSerializer(
                    lambda state_block_number: hex(state_block_number)
                    if isinstance(state_block_number, int)
                    else state_block_number,
                    return_type=HexStr,
                    when_used="json",
                ),
            ] = pydantic.Field(serialization_alias="stateBlockNumber")
            timestamp: int | None = None

        class EthCallBundlePayload(pydantic.BaseModel):
            jsonrpc: str = "2.0"
            id: int = 1
            method: str = "eth_callBundle"
            params: list[EthCallBundleParams]

        endpoint_method = "eth_callBundle"

        if endpoint_method not in self.endpoints:
            raise DegenbotValueError(
                message=f"{endpoint_method} was not included in the list of supported endpoints."
            )

        bundle_params = EthCallBundleParams(
            txs=[tx.to_0x_hex() for tx in bundle],
            block=block_number,
            state_block=state_block,
            timestamp=block_timestamp,
        )

        payload = EthCallBundlePayload(params=[bundle_params]).model_dump_json(by_alias=True)

        bundle_headers = self._build_authentication_header(
            payload=payload,
            header_label=self.authentication_header_label,
            signer_key=signer_key,
        )

        return await self.send_payload(
            url=self.url,
            http_session=http_session,
            headers=bundle_headers,
            data=payload,
        )

    async def cancel_eth_bundle(
        self,
        uuid: str | None = None,
        signer_key: str | None = None,
        http_session: aiohttp.ClientSession | None = None,
    ) -> Any:
        """
        Send a cancellation for the specified UUID.
        """

        endpoint_method = "eth_cancelBundle"

        if endpoint_method not in self.endpoints:
            raise DegenbotValueError(
                message=f"{endpoint_method} was not included in the list of supported endpoints."
            )

        if self.authentication_header_label is not None and signer_key is None:
            raise DegenbotValueError(
                message=f"Must provide signing address and key for required header {self.authentication_header_label}"  # noqa: E501
            )

        payload = json.dumps(
            {
                "jsonrpc": "2.0",
                "id": 1,
                "method": endpoint_method,
                "params": [
                    {
                        "replacementUuid": uuid,
                    }
                ],
            }
        )

        bundle_headers = self._build_authentication_header(
            payload=payload,
            signer_key=signer_key,
            header_label=self.authentication_header_label,
        )

        return await self.send_payload(
            url=self.url,
            http_session=http_session,
            headers=bundle_headers,
            data=payload,
        )

    async def cancel_private_transaction(
        self,
        tx_hash: HexBytes | str,
        signer_key: str,
        http_session: aiohttp.ClientSession | None = None,
    ) -> Any:
        """
        Send a cancellation for the specified private transaction hash.

        The signer key used for the request must match the signer key for the transaction.
        """

        endpoint_method = "eth_cancelPrivateTransaction"

        if endpoint_method not in self.endpoints:
            raise DegenbotValueError(
                message=f"{endpoint_method} was not included in the list of supported endpoints."
            )

        if isinstance(tx_hash, HexBytes):
            tx_hash = tx_hash.to_0x_hex()

        payload = json.dumps(
            {
                "jsonrpc": "2.0",
                "id": 1,
                "method": endpoint_method,
                "params": [
                    {
                        "txHash": tx_hash,
                    }
                ],
            }
        )

        bundle_headers = self._build_authentication_header(
            payload=payload,
            signer_key=signer_key,
            header_label=self.authentication_header_label,
        )

        return await self.send_payload(
            url=self.url,
            http_session=http_session,
            headers=bundle_headers,
            data=payload,
        )

    async def get_user_stats(
        self,
        signer_key: str,
        recent_block_number: BlockNumber,
        http_session: aiohttp.ClientSession | None = None,
    ) -> Any:
        """
        Get the Flashbots V2 user stats for the given searcher identity.
        """

        endpoint_method = "flashbots_getUserStatsV2"

        if endpoint_method not in self.endpoints:
            raise DegenbotValueError(
                message=f"{endpoint_method} was not included in the list of supported endpoints."
            )

        payload = json.dumps(
            {
                "jsonrpc": "2.0",
                "id": 1,
                "method": endpoint_method,
                "params": [
                    {
                        "blockNumber": hex(recent_block_number),
                    }
                ],
            }
        )

        bundle_headers = self._build_authentication_header(
            payload=payload,
            signer_key=signer_key,
            header_label=self.authentication_header_label,
        )

        return await self.send_payload(
            url=self.url,
            http_session=http_session,
            headers=bundle_headers,
            data=payload,
        )

    async def get_bundle_stats(
        self,
        bundle_hash: str,
        block_number: BlockNumber,
        signer_key: str,
        http_session: aiohttp.ClientSession | None = None,
    ) -> Any:
        """
        Get the Flashbots V2 bundle stats for the given searcher identity.
        """

        endpoint_method = "flashbots_getBundleStatsV2"

        if endpoint_method not in self.endpoints:
            raise DegenbotValueError(
                message=f"{endpoint_method} was not included in the list of supported endpoints."
            )

        payload = json.dumps(
            {
                "jsonrpc": "2.0",
                "id": 1,
                "method": endpoint_method,
                "params": [
                    {
                        "bundleHash": bundle_hash,
                        "blockNumber": hex(block_number),
                    }
                ],
            }
        )

        bundle_headers = self._build_authentication_header(
            payload=payload,
            signer_key=signer_key,
            header_label=self.authentication_header_label,
        )

        return await self.send_payload(
            url=self.url,
            http_session=http_session,
            headers=bundle_headers,
            data=payload,
        )

    async def send_eth_bundle(
        self,
        bundle: Iterable[HexBytes],
        block_number: BlockNumber,
        min_timestamp: int | None = None,
        max_timestamp: int | None = None,
        reverting_hashes: list[str] | None = None,
        uuid: str | None = None,
        signer_key: str | None = None,
        http_session: aiohttp.ClientSession | None = None,
    ) -> Any:
        """
        Send a formatted bundle to the eth_sendBundle endpoint
        """

        endpoint_method = "eth_sendBundle"

        if endpoint_method not in self.endpoints:
            raise DegenbotValueError(
                message=f"{endpoint_method} was not included in the list of supported endpoints."
            )

        if self.authentication_header_label is not None and signer_key is None:
            raise DegenbotValueError(
                message=f"Must provide signing address and key for required header {self.authentication_header_label}"  # noqa:E501
            )

        bundle_params: dict[str, Any] = {
            "txs": (
                # Array[String], A list of signed transactions to execute in an atomic bundle
                [tx.to_0x_hex() for tx in bundle]
            ),
            "blockNumber": (
                # String, a hex encoded block number for which this bundle is valid on
                hex(block_number)
            ),
        }

        if min_timestamp is not None:
            bundle_params[
                # (Optional) Number, the minimum timestamp for which this bundle is valid, in
                # seconds since the unix epoch
                "minTimestamp"
            ] = min_timestamp

        if max_timestamp is not None:
            bundle_params[
                # (Optional) Number, the maximum timestamp for which this bundle is valid, in
                # seconds since the unix epoch
                "maxTimestamp"
            ] = max_timestamp

        if reverting_hashes is not None:
            bundle_params[
                # (Optional) Array[String], A list of tx hashes that are allowed to revert
                "revertingTxHashes"
            ] = reverting_hashes

        if uuid is not None:
            bundle_params[
                # (Optional) String, UUID that can be used to cancel/replace this bundle
                "replacementUuid"
            ] = uuid

        payload = json.dumps(
            {
                "jsonrpc": "2.0",
                "id": 1,
                "method": endpoint_method,
                "params": [bundle_params],
            }
        )

        bundle_headers = self._build_authentication_header(
            payload=payload,
            signer_key=signer_key,
            header_label=self.authentication_header_label,
        )

        return await self.send_payload(
            url=self.url,
            headers=bundle_headers,
            data=payload,
            http_session=http_session,
        )

    async def send_private_transaction(
        self,
        raw_transaction: HexBytes | str,
        signer_key: str,
        max_block_number: BlockNumber | None = None,
        preferences: dict[str, Any] | None = None,
        http_session: aiohttp.ClientSession | None = None,
    ) -> Any:
        """
        Send a private raw transaction.
        """

        endpoint_method = "eth_sendPrivateTransaction"

        if endpoint_method not in self.endpoints:
            raise DegenbotValueError(
                message=f"{endpoint_method} was not included in the list of supported endpoints."
            )

        if isinstance(raw_transaction, HexBytes):
            raw_transaction = raw_transaction.to_0x_hex()

        params_dict: dict[str, Any] = {
            "tx": raw_transaction,
        }

        if max_block_number is not None:
            params_dict["maxBlockNumber"] = max_block_number

        if preferences is not None:
            params_dict["preferences"] = preferences

        payload = json.dumps(
            {
                "jsonrpc": "2.0",
                "id": 1,
                "method": endpoint_method,
                "params": [
                    params_dict,
                ],
            }
        )

        bundle_headers = self._build_authentication_header(
            payload=payload,
            signer_key=signer_key,
            header_label=self.authentication_header_label,
        )

        return await self.send_payload(
            url=self.url,
            http_session=http_session,
            headers=bundle_headers,
            data=payload,
        )

    async def send_private_raw_transaction(
        self,
        raw_transaction: HexBytes | str,
        signer_key: str,
        preferences: dict[str, Any] | None = None,
        http_session: aiohttp.ClientSession | None = None,
    ) -> Any:
        """
        Send a private raw transaction.
        """

        endpoint_method = "eth_sendPrivateRawTransaction"

        if endpoint_method not in self.endpoints:
            raise DegenbotValueError(
                message=f"{endpoint_method} was not included in the list of supported endpoints."
            )

        if isinstance(raw_transaction, HexBytes):
            raw_transaction = raw_transaction.to_0x_hex()

        params: list[str | dict[str, Any]] = [
            raw_transaction,
        ]

        if preferences is not None:
            params.append(preferences)

        payload = json.dumps(
            {
                "jsonrpc": "2.0",
                "id": 1,
                "method": endpoint_method,
                "params": params,
            }
        )

        bundle_headers = self._build_authentication_header(
            payload=payload,
            signer_key=signer_key,
            header_label=self.authentication_header_label,
        )

        return await self.send_payload(
            url=self.url,
            http_session=http_session,
            headers=bundle_headers,
            data=payload,
        )
